#
# Created : 2018-04-04
#
# @author: Eric Lapouyade
#
import copy
import re
from datetime import timedelta
from itertools import chain

from dal import autocomplete
from django import forms
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.db.models import QuerySet
from django.forms import FileField, CheckboxInput
from django.template import loader
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy, pgettext_lazy

from . import FILTER_QUERYSTRING_PREFIX
from .context import RenderContext
from .exceptions import InvalidFilters
from .form_fields import ListOfValuesField
from .html_attributes import HTMLAttributes
from .theme_config import ThemeTemplate, ThemeAttribute
from .utils import init_dicts_from_class

__all__ = [
    "Filter",
    "AutocompleteForeignKeyFilter",
    "AutocompleteMultipleForeignKeyFilter",
    "BooleanFilter",
    "ChoiceFilter",
    "DateFilter",
    "DateTimeFilter",
    "FILTER_QUERYSTRING_PREFIX",
    "Filters",
    "FILTERS_PARAMS_KEYS",
    "ForeignKeyFilter",
    "IntegerFilter",
    "FloatFilter",
    "MultipleChoiceFilter",
    "MultipleForeignKeyFilter",
    "TimeFilter",
    "NoFilter",
    "DEFAULT_FILTER_HELP_TEXT",
]

# Declare keys only for "Filters" object
FILTERS_KEYS = {
    "form_attrs",
    "form_buttons",
    "form_layout",
    "form_layout_advanced",
    "form_reset_label",
    "form_submit_label",
    "form_advanced_label",
    "form_template_name",
    "theme_form_submit_icon",
    "theme_form_reset_icon",
    "theme_form_submit_class",
    "theme_form_reset_class",
}

# Declare keys for "Filter" object (not "Filters" with an ending 's')
FILTERS_PARAMS_KEYS = {
    "format_label",
    "container_attrs",
    "filter_distinct",
    "filter_exclude",
    "filter_key",
    "field_name",
    "input_type",
    "key_type",
    "model_field",
    "name",
    "no_choice_msg",
    "no_filtering",
    "order_by",
    "queryset",
    "shrink_width",
    "flex_width",
    "widget_attrs",
    "widget_class",
    "widget_params",
    "word_search",
    "url",
    "choices",
    "filter_queryset_method",
    "add_one_day",
    "auto_help_text",
    "default_value",
    "default_value_func",
    "form_field_input_container_css",
    "reverse_form_label_tag",
    "resolve_qs_value",
}

# Declare keys for django form fields
FILTERS_FORM_FIELD_KEYS = {
    "choices",
    "disabled",
    "empty_value",
    "empty_label",
    "queryset",
    "error_messages",
    "help_text",
    "initial",
    "label",
    "label_suffix",
    "localize",
    "max_length",
    "min_length",
    "required",
    "strip",
    "validators",
    "widget",
}

FILTERS_PARAMS_KEYS.update(FILTERS_KEYS)
FILTERS_PARAMS_KEYS.update(FILTERS_FORM_FIELD_KEYS)
DEFAULT_FILTER_HELP_TEXT = "&nbsp;"


class FiltersBaseForm(forms.BaseForm):
    def _clean_fields(self):
        for name, field in self.fields.items():
            if field.disabled:
                value = self.get_initial_for_field(field, name)
            else:
                value = field.widget.value_from_datadict(
                    self.data, self.files, self.add_prefix(name)
                )
                if not value and field.required:
                    value = self.get_initial_for_field(field, name)
                    self.data[name] = value

            try:
                if isinstance(field, FileField):
                    initial = self.get_initial_for_field(field, name)
                    value = field.clean(value, initial)
                else:
                    value = field.clean(value)
                self.cleaned_data[name] = value

                if value is None and not field.required:
                    continue
                filter_name = name[len(FILTER_QUERYSTRING_PREFIX) :]
                method_name = f"filter_form_clean_{filter_name}"
                method = getattr(self.listing, method_name, None)
                if method:
                    value = method(self)
                    self.cleaned_data[name] = value
            except ValidationError as e:
                self.add_error(name, e)

    def _clean_form(self):
        try:
            method_name = f"filter_form_clean"
            method = getattr(self.listing, method_name, None)
            if method:
                cleaned_data = method(self)
            else:
                cleaned_data = self.cleaned_data
        except ValidationError as e:
            self.add_error(None, e)
        else:
            if cleaned_data is not None:
                self.cleaned_data = cleaned_data


class Filters(list):
    id = None
    listing = None
    css_id = None
    form_reset_label = pgettext_lazy("Filters form", "Reset")
    form_submit_label = pgettext_lazy("Filters form", "Filter")
    form_advanced_label = pgettext_lazy("Filters form", "Advanced")
    form_template_name = ThemeTemplate("filters_form.html")
    theme_form_reset_icon = ThemeAttribute("filters_theme_form_reset_icon")
    theme_form_reset_class = ThemeAttribute("filters_theme_form_reset_class")
    theme_form_submit_icon = ThemeAttribute("filters_theme_form_submit_icon")
    theme_form_submit_class = ThemeAttribute("filters_theme_form_submit_class")
    theme_form_advanced_down_icon = ThemeAttribute(
        "filters_theme_form_advanced_down_icon"
    )
    theme_form_advanced_up_icon = ThemeAttribute("filters_theme_form_advanced_up_icon")
    theme_form_advanced_class = ThemeAttribute("filters_theme_form_advanced_class")
    form_layout = None
    form_layout_advanced = None
    form_buttons = "reset,submit"
    show_advanced = False

    def __init__(self, *filters, params=None, **kwargs):
        self.init_kwargs = kwargs
        # params contains Filter objects parameters, ex :
        # params = {"filter1": {"param1" : "value1" ...}, "filter2": {"param2" : "value2" ...}
        if params is None:
            params = {}
        self._params = params
        self._form = None
        self._cleaned_data = None
        self.name2filter = {}
        init_dicts_from_class(self, ["form_attrs"])
        super().__init__(filters)

    def get(self, name, default=None):
        return self.name2filter.get(name, default)

    def get_params(self):
        return self._params

    def names(self):
        return self.name2filter.keys()

    def select(self, select_filters_name=None, exclude_filters_name=None):
        if isinstance(select_filters_name, str):
            select_filters_name = map(str.strip, select_filters_name.split(","))
        if isinstance(exclude_filters_name, str):
            exclude_filters_name = map(str.strip, exclude_filters_name.split(","))
        select_filters_name = select_filters_name or self.names()
        exclude_filters_name = exclude_filters_name or []
        return [
            self.get(c.strip())
            for c in select_filters_name
            if c not in exclude_filters_name
        ]

    def normalize_layout(self, layout):
        if isinstance(layout, str):
            # transform layout string into list of lists of lists
            layout = re.sub(r"\s", "", layout)
            layout = list(
                map(
                    lambda s: list(
                        map(
                            lambda t: (t.split("|") + [None, None, None])[:4],
                            filter(None, s.split(",")),
                        )
                    ),
                    filter(None, layout.split(";")),
                )
            )
        if layout is None:
            layout = []
        return layout

    def bind_to_listing(self, listing):
        filters = self.__class__(params=self._params, **self.init_kwargs)
        for k in FILTERS_KEYS:
            if k in self.init_kwargs:
                setattr(filters, k, self.init_kwargs[k])
            elif hasattr(self, k):
                setattr(filters, k, getattr(self, k))
            listing_key = "filters_" + k
            if hasattr(listing, listing_key):
                setattr(filters, k, getattr(listing, listing_key))
        if filters.form_layout is None:
            filters.form_layout = [
                [[filtr.name, None, None, None]]
                for filtr in filters
                if isinstance(filtr, Filter)
            ]
        filters.form_layout = self.normalize_layout(filters.form_layout)
        filters.form_layout_advanced = self.normalize_layout(
            filters.form_layout_advanced
        )
        for filtr in self:
            if not isinstance(filtr, Filter):
                # understand, filtr is a string
                filtr = self.create_filter(filtr, listing)
            # ensure there is one filter instance per listing
            elif not filtr.listing:
                filtr = copy.copy(filtr)  # do not use deepcopy here
            if isinstance(filtr, Filter):
                filtr.bind_to_listing(listing)
                filters.append(filtr)
        filters.name2filter = {f.name: f for f in filters if isinstance(f, Filter)}
        filters.listing = listing
        # extract labels and endings from given layout
        for row in chain(filters.form_layout, filters.form_layout_advanced):
            for name, label, ending, input_type in row:
                filtr = filters.name2filter.get(name)
                if filtr:
                    if label:
                        filtr.label = label
                    if ending:
                        filtr.ending = ending
                    if input_type:
                        filtr.input_type = input_type
        fbuttons = filters.form_buttons
        if isinstance(fbuttons, str):
            filters.form_buttons = list(map(str.strip, fbuttons.split(",")))
        return filters

    def datetimepicker_init(self):
        if self.listing.use_datetimepicker:
            self.listing.need_media_for("datetimepicker")
            self.listing.add_onready_snippet(
                f"""
                $('#{self.id} .edit-datecolumn').datetimepicker({{timepicker:false, format:'{self.listing.datetimepicker_date_format}'}});
                $('#{self.id} .edit-datetimecolumn').datetimepicker({{format:'{self.listing.datetimepicker_datetime_format}'}});
                $('#{self.id} .edit-timecolumn').datetimepicker({{datepicker:false, format:'{self.listing.datetimepicker_time_format}'}});
                """
            )

    def create_filter(self, name, listing):
        if isinstance(name, str):
            model_attr_name, *dummy = name.split("__")
            if issubclass(listing.model, models.Model):
                field = listing.model._meta.get_field(model_attr_name)
                if field:
                    for filter_class in FilterMeta.get_filter_classes():
                        if filter_class is not Filter:
                            filtr = filter_class.from_model_field(name, field)
                            if filtr:
                                return filtr
            return Filter(name)
        return None

    def extract_params(self):
        request_data = self.listing.request_data
        for f in self:
            if not request_data and (
                f.default_value is not None or callable(f.default_value_func)
            ):
                f.value = f.get_default_value()
            else:
                f.extract_params(request_data)

    def get_cleaned_data(self):
        # Cache result to self._cleaned_data,
        # This is possible because self is a copy of the Filters object
        # defined in the listing class. Copy is done on every request
        if not self._cleaned_data:
            form = self.form()
            self._cleaned_data = form.cleaned_data if form.is_valid() else None
            if self._cleaned_data is None:
                self._cleaned_data = self.get_default_cleaned_data()
        return self._cleaned_data

    def filter_queryset(self, qs):
        cleaned_data = self.get_cleaned_data()
        for filtr in self:
            qs = filtr.filter_queryset(qs, cleaned_data)
        return qs

    def get_default_cleaned_data(self):
        return {
            (filtr.input_name + filtr.listing.suffix): filtr.get_default_value()
            for filtr in self
        }

    def form(self):
        # Cache result to self._form,
        # This is possible because self is a copy of the Filters object
        # defined in the listing class. Copy is done on every request
        if not self._form:
            fields = {
                f.input_name + self.listing.suffix: f.create_form_field() for f in self
            }
            initial = {
                f.input_name + self.listing.suffix: f.default_value_func(f)
                if callable(f.default_value_func)
                else f.default_value
                for f in self
                if f.default_value is not None or callable(f.default_value_func)
            }
            form_class = type(
                "FilterForm{}".format(self.listing.suffix),
                (FiltersBaseForm,),
                {"base_fields": fields},
            )
            if self.listing.request.method == "POST":
                for field_name, field in fields.items():
                    if isinstance(field.widget, CheckboxInput):
                        if field_name not in self.listing.request.POST:
                            # Need to do that because
                            # checkbox does not return off values
                            if field_name in self.listing.request_data:
                                del self.listing.request_data[field_name]
            self._form = form_class(self.listing.request_data or None, initial=initial)
            self._form.listing = self.listing
        return self._form

    def get_hiddens_html(self):
        query_fields = [(FILTER_QUERYSTRING_PREFIX + f.name) for f in self]
        query_fields.append(FILTER_QUERYSTRING_PREFIX + "do_filter")
        return self.listing.get_hiddens_html(without=query_fields)

    def render_init(self, context):
        self.listing.manage_page_context(context)
        advanced_filters_names = {b[0] for a in self.form_layout_advanced for b in a}
        requested_filter = {
            k[len(FILTER_QUERYSTRING_PREFIX) :]
            for k, v in self.listing.request.GET.items()
            if k.startswith(FILTER_QUERYSTRING_PREFIX) and v
        }
        if advanced_filters_names & requested_filter:
            self.show_advanced = True
        if not isinstance(self.form_attrs, HTMLAttributes):
            self.form_attrs = HTMLAttributes(self.form_attrs)
        if "method" not in self.form_attrs:
            self.form_attrs.add("method", "POST")
        self.form_attrs.add("class", {"listing-form", "filters-form"})
        if self.listing.accept_ajax:
            self.form_attrs.add("class", "filters-form-ajax")
        if "id" not in self.form_attrs:
            self.form_attrs.add("id", "filters-form-id{}".format(self.listing.suffix))
        self.css_id = self.form_attrs.get("id")
        self.form_attrs.add("listing-id", self.listing.css_id)
        self.id = self.form_attrs["id"]
        self.datetimepicker_init()

    def render_form(self, context):
        self.render_init(context)
        ctx = self.get_context()
        template = loader.get_template(self.form_template_name)
        out = template.render(ctx)
        return out

    def get_form_reset_url(self):
        query_fields = [(FILTER_QUERYSTRING_PREFIX + f.name) for f in self]
        query_fields += [f"{FILTER_QUERYSTRING_PREFIX}do_filter"]
        return self.listing.get_url(without=query_fields)

    def get_form_field(self, name):
        filter = self.get(name)
        if filter is None:
            raise InvalidFilters(
                _('Cannot find filter "{name}" definition').format(name=name)
            )
        return self.form()[filter.input_name + self.listing.suffix]

    def get_filter_value(self, name):
        key = f"{FILTER_QUERYSTRING_PREFIX}{name}"
        form = self.form()
        if "cleaned_data" in form:
            return form.cleaned_data.get(key)
        return form.initial.get(key)

    def get_form_field_container_attrs(self, name):
        filter = self.get(name)
        if filter is None:
            raise InvalidFilters(
                _('Cannot find filter "{name}" definition').format(name=name)
            )
        form_field = self.form()[filter.input_name + self.listing.suffix]
        return filter.get_form_field_container_attrs(form_field)

    def get_context(self):
        ctx = RenderContext(
            self.listing.global_context,
            self.listing.page_context.flatten(),
            listing=self.listing,
            filters=self,
            get=self.listing.request.GET,
        )
        return ctx


class FilterMeta(type):
    filter_classes = []

    def __new__(mcs, name, bases, attrs):
        cls = super(FilterMeta, mcs).__new__(mcs, name, bases, attrs)
        FilterMeta.filter_classes.append((cls, cls.from_model_field_order))
        params_keys = cls.params_keys
        if params_keys:
            if isinstance(params_keys, str):
                params_keys = set(map(str.strip, params_keys.split(",")))
            FILTERS_PARAMS_KEYS.update(params_keys)
            FILTERS_PARAMS_KEYS.discard("")
        form_field_keys = cls.form_field_keys
        if form_field_keys:
            if isinstance(form_field_keys, str):
                form_field_keys = set(map(str.strip, form_field_keys.split(",")))
            FILTERS_FORM_FIELD_KEYS.update(form_field_keys)
            FILTERS_FORM_FIELD_KEYS.discard("")
            FILTERS_PARAMS_KEYS.update(FILTERS_FORM_FIELD_KEYS)
        return cls

    @classmethod
    def get_filter_classes(cls):
        return [t[0] for t in sorted(cls.filter_classes, key=lambda x: x[1])]


class Filter(metaclass=FilterMeta):
    container_attrs = {"class": "form-field"}
    default_value = None
    default_value_func = None
    field_name = None
    filter_distinct = False
    filter_exclude = False
    filter_key = None
    filter_queryset_method = None
    flex_width = None
    form_field_class = forms.CharField
    form_field_keys = None
    format_label = None
    from_model_field_classes = ()
    from_model_field_name = None  # used by GroupByFilterColumn indirectly
    from_model_field_order = 100
    help_text = None
    input_name = None
    input_type = None
    key_type = None
    label = None
    listing = None
    name = None
    no_choice_msg = gettext_lazy("- No filtering -")
    no_filtering = False
    order_by = None
    params_keys = None
    queryset = None
    required = False
    resolve_qs_value = False
    shrink_width = None
    url = None
    value = None
    widget_attrs = None
    widget_class = None
    widget_params = None
    word_search = False
    form_field_input_container_css = ""
    reverse_form_label_tag = False

    theme_form_widget_class = ThemeAttribute("column_theme_form_widget_class")
    theme_form_select_widget_class = ThemeAttribute(
        "column_theme_form_select_widget_class"
    )
    theme_form_checkbox_widget_class = ThemeAttribute(
        "column_theme_form_checkbox_widget_class"
    )
    theme_form_radio_widget_class = ThemeAttribute(
        "column_theme_form_radio_widget_class"
    )

    def __init__(self, *args, **kwargs):
        self.init_args = args
        self.init_kwargs = kwargs
        init_dicts_from_class(
            self, ["widget_attrs", "form_field_attrs", "widget_params"]
        )
        if isinstance(self.theme_form_widget_class, str):
            self.theme_form_widget_class = set(self.theme_form_widget_class.split())
        if isinstance(self.theme_form_select_widget_class, str):
            self.theme_form_select_widget_class = set(
                self.theme_form_select_widget_class.split()
            )
        if isinstance(self.theme_form_checkbox_widget_class, str):
            self.theme_form_checkbox_widget_class = set(
                self.theme_form_checkbox_widget_class.split()
            )
        if isinstance(self.theme_form_radio_widget_class, str):
            self.theme_form_radio_widget_class = set(
                self.theme_form_radio_widget_class.split()
            )

    def bind_to_listing(self, listing):
        self.init(listing, *self.init_args, **self.init_kwargs)

    def set_listing(self, listing):
        self.listing = listing

    def init(self, listing, name=None, **kwargs):
        self.set_listing(listing)
        if name:
            self.name = re.sub(r"\W", "", name)
        self.set_kwargs(**kwargs)
        self.apply_template_kwargs()
        if self.filter_key is None:
            self.filter_key = self.name
        self.filter_key = self.filter_key.replace(".", "__")
        if self.field_name is None:
            self.field_name, *dummy = self.filter_key.split("__")
        if self.format_label is None:
            self.format_label = lambda obj: str(obj)
        elif isinstance(self.format_label, str):
            field = str(self.format_label)
            self.format_label = lambda obj: getattr(obj, field, "-")
        if self.input_name is None:
            self.input_name = FILTER_QUERYSTRING_PREFIX + self.name
        self.label = self.get_label()
        if self.help_text is None:
            self.help_text = DEFAULT_FILTER_HELP_TEXT  # otherwise formfields may me not aligned vertically. keep align-items: flex-end; on row
        if self.from_model_field_name is None:
            self.from_model_field_name, *dummy = self.filter_key.split("__")

    def get_label(self):
        label = self.label
        if label is None:
            try:
                label = self.listing.model._meta.get_field(self.field_name).verbose_name
            except FieldDoesNotExist:
                pass
        if label is None:
            label = self.field_name.replace("_", " ").capitalize()
        return label

    def get_form_field_container_attrs(self, form_field):
        attrs = HTMLAttributes(self.container_attrs)
        attrs.add(
            "class",
            {
                "col-" + self.name,
                "cls-" + self.__class__.__name__.lower(),
                # self.listing.filters.theme_field_class,
            },
        )
        if self.shrink_width:
            attrs.add("style", f"flex-shrink: {self.shrink_width}")
        if self.flex_width:
            attrs.add("style", f"flex: {self.flex_width}")
        if form_field.errors:
            attrs.add("class", {"errors"})
        return attrs

    def set_kwargs(self, **kwargs):
        # If filters_<filterclass>_<params> exists in listing attributes :
        # take is as a default value
        filter_class_slug = self.__class__.__name__.lower()
        for k in FILTERS_PARAMS_KEYS:
            filter_key = "filters_{}".format(k)
            if hasattr(self.listing, filter_key):
                setattr(self, k, getattr(self.listing, filter_key))
            filter_key = "{}s_{}".format(filter_class_slug, k)
            if hasattr(self.listing, filter_key):
                setattr(self, k, getattr(self.listing, filter_key))
        for k, v in kwargs.items():
            if k in FILTERS_PARAMS_KEYS:
                setattr(self, k, v)
        # if parameters given in filters : apply them
        for k, v in self.listing.filters.get_params().get(self.name, {}).items():
            if k in FILTERS_PARAMS_KEYS:
                setattr(self, k, v)
        # if parameters given in listing apply them to specified filter
        for k in FILTERS_PARAMS_KEYS:
            key = "{}{}__{}".format(FILTER_QUERYSTRING_PREFIX, self.name, k)
            v = getattr(self.listing, key, None)
            if v is not None:
                setattr(self, k, v)

    def apply_template_kwargs(self):
        kwargs = self.listing.get_filter_kwargs(self.name)
        if kwargs:
            for k, v in kwargs.items():
                if k in FILTERS_PARAMS_KEYS:
                    if isinstance(v, dict):
                        prev_value = getattr(self, k, None)
                        if isinstance(prev_value, dict):
                            prev_value.update(v)
                            continue
                    setattr(self, k, v)

    @classmethod
    def from_model_field(cls, name, field):
        if field.__class__ in cls.from_model_field_classes:
            return cls(name, model_field=field, label=field.verbose_name)

    def get_form_field_params(self, **kwargs):
        params = {
            p: getattr(self, p)
            for p in FILTERS_FORM_FIELD_KEYS
            if getattr(self, p, None) is not None
        }
        return params

    def get_form_field_class(self, **kwargs):
        return self.form_field_class

    def get_form_field_widget(self, field_class, **kwargs):
        widget_class = self.widget_class or field_class.widget

        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        widget_attrs.add("class", self.theme_form_widget_class)
        widget_id = f"id-filter-{self.name}{self.listing.suffix}".replace("_", "-")
        widget_attrs.add("id", widget_id)
        params = dict(self.widget_params, attrs=widget_attrs)
        return widget_class(**params)

    def create_form_field(self):
        cls = self.get_form_field_class()
        params = self.get_form_field_params()
        widget = self.get_form_field_widget(cls)
        if self.listing.model and self.from_model_field_name:
            try:
                model_field = self.listing.model._meta.get_field(
                    self.from_model_field_name
                )
                widget.attrs["data-model-field"] = model_field.name
                if related_model := getattr(model_field, "related_model", None):
                    if not widget.attrs.get("data-related-model"):
                        widget.attrs["data-related-model"] = (
                            f"{related_model._meta.app_label}"
                            f".{related_model._meta.model_name}"
                        )
            except FieldDoesNotExist:
                pass
        field = cls(widget=widget, **params)
        patch_help_text_method = getattr(widget, "patch_help_text", None)
        if callable(patch_help_text_method):
            field.help_text = patch_help_text_method(field.help_text)

        return field

    def get_default_value(self):
        default_value = (
            self.default_value_func(self)
            if callable(self.default_value_func)
            else self.default_value
        )
        return default_value

    def get_queryset_clean_value(self, cleaned_data):
        cleaned_value = cleaned_data.get(self.input_name + self.listing.suffix)
        return cleaned_value

    def filter_queryset(self, qs, cleaned_data=None):
        if self.no_filtering:
            return qs
        cleaned_value = self.get_queryset_clean_value(cleaned_data)
        if self.resolve_qs_value and isinstance(cleaned_value, QuerySet):
            # when using multi-selection select, Django returns the queryset
            # representing the selection : this may cause very slow filtering.
            # Instead, using PKs corresponding to this selection are much faster :
            # this is the purpose of resolve_qs_value attribute
            cleaned_value = list(cleaned_value.values_list("pk", flat=True))
        if cleaned_value is not False and not cleaned_value:
            return qs
        method_name = f"filter_queryset_{self.name}"
        method = getattr(self.listing, method_name, None)
        if not method and self.filter_queryset_method:
            if isinstance(self.filter_queryset_method, str):
                method = getattr(self.listing, self.filter_queryset_method, None)
            else:
                method = self.filter_queryset_method
        if method:
            if self.word_search and isinstance(cleaned_value, str):
                words = filter(None, cleaned_value.split())
                for word in words:
                    qs = method(qs, word)
                return qs
            return method(qs, cleaned_value)
        if self.filter_exclude:
            filter_op = qs.exclude
        else:
            filter_op = qs.filter
        if self.word_search and isinstance(cleaned_value, str):
            words = filter(None, cleaned_value.split())
            for word in words:
                # DO NOT use filter_op here, not working in loops
                qs = qs.filter(**{self.filter_key: word})
        elif "__" not in self.filter_key and isinstance(
            cleaned_value, (QuerySet, list, tuple)
        ):
            qs = filter_op(**{f"{self.filter_key}__in": cleaned_value})
        else:
            qs = filter_op(**{self.filter_key: cleaned_value})
        if self.filter_distinct:
            qs = qs.distinct()
        return qs

    def filter_sequence(self, seq):
        if not self.value:
            return seq
        key_filter = (
            self.filter_key if "__" in self.filter_key else f"{self.filter_key}__equal"
        )
        key, filtr = key_filter.split("__", 1)
        if filtr == "equal":
            return filter(lambda rec: rec[key] == self.value, seq)
        elif filtr == "contains":
            return filter(
                lambda rec: rec[key] is not None and self.value in str(rec[key]), seq
            )
        elif filtr == "icontains":
            return filter(
                lambda rec: rec[key] is not None
                and self.value.lower() in str(rec[key]).lower(),
                seq,
            )
        elif filtr == "regex":
            return filter(lambda rec: re.search(self.value, rec[key]), seq)
        elif filtr == "iregex":
            return filter(lambda rec: re.search(self.value, rec[key], flags=re.I), seq)
        elif filtr == "lt":
            return filter(
                lambda rec: rec[key] is not None and rec[key] < self.value, seq
            )
        elif filtr == "lte":
            return filter(
                lambda rec: rec[key] is not None and rec[key] <= self.value, seq
            )
        elif filtr == "gt":
            return filter(
                lambda rec: rec[key] is not None and rec[key] > self.value, seq
            )
        elif filtr == "gte":
            return filter(
                lambda rec: rec[key] is not None and rec[key] >= self.value, seq
            )
        return seq

    def extract_params(self, request_get_data):
        self.value = request_get_data.get(self.input_name + self.listing.suffix)

    def set_params_choices(self, params):
        if not params.get("choices"):
            if hasattr(self, "model_field"):
                params["choices"] = self.model_field.choices
            elif self.listing.model:
                try:
                    model = self.listing.model
                    params["choices"] = model._meta.get_field(self.field_name).choices
                except AttributeError:
                    raise InvalidFilters(
                        _(
                            "Cannot find choices from model "
                            'for filter "{name}", Please specify the choices to '
                            "display"
                        ).format(name=self.name)
                    )
            else:
                raise InvalidFilters(
                    _(
                        "Please specify the choices to display " 'for filter "{name}"'
                    ).format(name=self.name)
                )


class IntegerFilter(Filter):
    from_model_field_classes = (models.IntegerField,)
    form_field_class = forms.IntegerField


class FloatFilter(Filter):
    from_model_field_classes = (models.FloatField,)
    form_field_class = forms.FloatField


class ListOfValuesFilter(Filter):
    form_field_class = ListOfValuesField

    def init(self, listing, name=None, **kwargs):
        super().init(listing, name, **kwargs)
        if not self.filter_key.endswith("__in"):
            self.filter_key += "__in"


class DateFilter(Filter):
    from_model_field_classes = (models.DateField,)
    form_field_class = forms.DateField
    widget_attrs = {"class": "form-control edit-datecolumn", "type": "date"}
    widget_params = {"format": "%Y-%m-%d"}
    add_one_day = False  # useful for DateFilter on DateTimeField
    auto_help_text = True

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        if self.auto_help_text and params.get("help_text") == DEFAULT_FILTER_HELP_TEXT:
            if self.filter_key.endswith("__gte"):
                params["help_text"] = _("Greater than or equal to")
            elif self.filter_key.endswith("__gt"):
                params["help_text"] = _("Greater than to")
            elif self.filter_key.endswith("__lte"):
                params["help_text"] = _("Less than or equal to")
            elif self.filter_key.endswith("__lt"):
                params["help_text"] = _("Less than to")
        return params

    def filter_queryset(self, qs, cleaned_data):
        if not self.value:
            return qs
        if self.add_one_day:
            data_key = self.input_name + self.listing.suffix
            cleaned_value = cleaned_data.get(data_key)
            if cleaned_value:
                cleaned_data[data_key] = cleaned_value + timedelta(days=1)
        return super().filter_queryset(qs, cleaned_data)


class DateTimeFilter(Filter):
    from_model_field_classes = (models.DateTimeField,)
    form_field_class = forms.DateTimeField
    widget_attrs = {
        "class": "form-control edit-datetimecolumn",
        "type": "datetime-local",
    }


class TimeFilter(Filter):
    from_model_field_classes = (models.TimeField,)
    form_field_class = forms.TimeField
    widget_attrs = {"class": "form-control edit-timecolumn", "type": "time"}


class BooleanFilter(Filter):
    from_model_field_classes = (models.BooleanField,)
    form_field_class = forms.ChoiceField
    true_msg = gettext_lazy("Yes")
    false_msg = gettext_lazy("No")
    indifferent_msg = gettext_lazy("Indiff.")

    def get_form_field_widget(self, field_class, force_select=False, **kwargs):
        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        if self.input_type == "radio" and not force_select:
            widget = forms.RadioSelect
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("class", "multiple-radios")
        elif self.input_type == "radioinline" and not force_select:
            widget = forms.RadioSelect
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("class", "multiple-radios inline")
        elif self.input_type == "switch" and not force_select:
            widget = forms.CheckboxInput
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("role", "switch")
        else:
            widget = forms.Select
            widget_attrs.add("class", self.theme_form_select_widget_class)
        params = dict(self.widget_params, attrs=widget_attrs)
        return widget(**params)

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        params["choices"] = [
            ("", self.indifferent_msg),
            ("True", self.true_msg),
            ("False", self.false_msg),
        ]
        return params

    def filter_queryset(self, qs, cleaned_data):
        data_key = self.input_name + self.listing.suffix
        cleaned_value = cleaned_data.get(data_key)
        if cleaned_value:
            cleaned_data[data_key] = cleaned_value is True or (
                isinstance(cleaned_value, str) and cleaned_value.lower() == "true"
            )
        return super().filter_queryset(qs, cleaned_data)


class ChoiceFilter(Filter):
    form_field_class = forms.ChoiceField
    form_field_keys = "choices"

    @classmethod
    def from_model_field(cls, name, field):
        choices = getattr(field, "choices", None)
        if choices:
            return cls(
                name, model_field=field, label=field.verbose_name, choices=choices
            )

    def get_form_field_widget(self, field_class, force_select=False, **kwargs):
        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        if self.input_type == "radio" and not force_select:
            widget = forms.RadioSelect
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("class", "multiple-radios")
        elif self.input_type == "radioinline" and not force_select:
            widget = forms.RadioSelect
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("class", "multiple-radios inline")
        else:
            widget = forms.Select
            widget_attrs.add("class", self.theme_form_select_widget_class)
        params = dict(self.widget_params, attrs=widget_attrs)
        return widget(**params)

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        self.set_params_choices(params)
        if callable(params["choices"]):
            params["choices"] = params["choices"](self)
        params["choices"] = [("", self.no_choice_msg)] + list(params["choices"])
        return params


class MultipleChoiceFilter(Filter):
    form_field_class = forms.MultipleChoiceField
    form_field_keys = "choices"
    from_model_field_order = 90
    # 90 < 100 (default) means it will consider MultipleChoiceFilter before
    # ChoiceFilter when searching the right filter for a model field.

    def init(self, listing, name=None, **kwargs):
        super().init(listing, name, **kwargs)
        if not self.filter_key.endswith("__in"):
            self.filter_key += "__in"

    @classmethod
    def from_model_field(cls, name, field):
        choices = getattr(field, "choices", None)
        if choices and name.endswith("__in"):
            return cls(name, model_field=field, label=field.verbose_name)

    def get_form_field_widget(self, field_class, **kwargs):
        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        if self.input_type == "checkbox":
            widget = forms.CheckboxSelectMultiple
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("class", "multiple-checkboxes")
        elif self.input_type == "checkboxinline":
            widget = forms.CheckboxSelectMultiple
            widget_attrs.add("class", self.theme_form_radio_widget_class)
            widget_attrs.add("class", "multiple-checkboxes inline")
        else:
            widget = forms.CheckboxInput
            widget_attrs.add("class", self.theme_form_select_widget_class)
        params = dict(self.widget_params, attrs=widget_attrs)
        return widget(**params)

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        self.set_params_choices(params)
        if params.get("help_text") is None and self.input_type not in [
            "checkbox",
            "checkboxinline",
        ]:
            params["help_text"] = _(
                "Use CTRL+click to select/deselect multiple choices"
            )
        return params

    def extract_params(self, request_get_data):
        self.value = request_get_data.getlist(self.input_name + self.listing.suffix)


class ForeignKeyFilter(Filter):
    form_field_class = forms.ChoiceField

    def get_form_field_widget(self, field_class, **kwargs):
        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        widget = forms.Select
        widget_attrs.add("class", self.theme_form_select_widget_class)
        params = dict(self.widget_params, attrs=widget_attrs)
        return widget(**params)

    def get_choices_order(self):
        return self.order_by

    def get_related_qs(self):
        if self.queryset is not None:
            return self.queryset
        related_model = self.listing.model
        for f_name in self.filter_key.split("__"):
            related_model = related_model._meta.get_field(f_name).related_model
        qs = related_model.objects.all()
        order_by = self.get_choices_order()
        if order_by:
            qs = qs.order_by(order_by)
        return qs

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        self.set_params_choices(params)
        choices = [("", self.no_choice_msg)]
        choices += [(obj.pk, self.format_label(obj)) for obj in self.get_related_qs()]
        params["choices"] = choices
        if "queryset" in params:
            del params["queryset"]
        return params


class MultipleForeignKeyFilter(Filter):
    form_field_class = forms.ModelMultipleChoiceField

    def get_form_field_widget(self, field_class, **kwargs):
        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        widget = forms.SelectMultiple
        widget_attrs.add("class", self.theme_form_select_widget_class)
        params = dict(self.widget_params, attrs=widget_attrs)
        return widget(**params)

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        params["queryset"] = self.queryset
        return params


class AutocompleteForeignKeyFilter(Filter):
    form_field_class = forms.ModelChoiceField
    widget_class = autocomplete.ModelSelect2

    def get_form_field_params(self, **kwargs):
        params = super().get_form_field_params(**kwargs)
        if "required" not in params:
            params["required"] = False
        if self.queryset is not None:
            params["queryset"] = self.queryset
        else:
            related_model = self.listing.model
            for f_name in self.filter_key.split("__"):
                if f_name == "in":
                    break
                try:
                    related_model = related_model._meta.get_field(f_name).related_model
                except FieldDoesNotExist:
                    break
            params["queryset"] = related_model.objects.all()
        return params

    def get_form_field_widget(self, field_class, **kwargs):
        self.listing.need_media_for("autocomplete")
        widget_attrs = self.widget_attrs or {}
        attrs = self.widget_params.get("attrs", {})
        widget_attrs = HTMLAttributes({**widget_attrs, **attrs})
        widget_attrs.add("class", self.theme_form_select_widget_class)
        if "data-placeholder" not in widget_attrs:
            widget_attrs["data-placeholder"] = _("Select a value...")
        widget = self.widget_class
        if self.url is None:
            raise InvalidFilters(
                f"Please specify the url name to autocomplete view for {self.name}"
            )
        if widget_attrs.get("data-html") is not False:
            widget_attrs["data-html"] = True
        params = dict(self.widget_params, url=self.url, attrs=widget_attrs)
        return widget(**params)


class AutocompleteMultipleForeignKeyFilter(AutocompleteForeignKeyFilter):
    form_field_class = forms.ModelMultipleChoiceField
    widget_class = autocomplete.ModelSelect2Multiple
    resolve_qs_value = True


class NoFilter(Filter):
    def filter_queryset(self, qs, cleaned_data=None):
        return qs



#
# Created : 2018-02-09
#
# @author: Eric Lapouyade
#
import copy
import re
from itertools import count
from types import GeneratorType

from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_listing import EXPORT_FORMATS

from .context import RenderContext
from .exceptions import *
from .html_attributes import HTMLAttributes
from .theme_config import ThemeAttribute, ThemeTemplate
from .utils import normalize_choices, normalize_list

__all__ = [
    "ExportDropdownToolbarItem",
    "ExportSelectToolbarItem",
    "PerPageDropdownToolbarItem",
    "PerPageSelectToolbarItem",
    "SortDropdownToolbarItem",
    "SortSelectToolbarItem",
    "TOOLBAR_PARAMS_KEYS",
    "Toolbar",
    "ToolbarItem",
    "UpdateToolbarItem",
    "VariationsToolbarItem",
    "VariationsDropdownToolbarItem",
    "GroupByToolbarItem",
    "LinkButtonToolbarItem",
]

TOOLBAR_PARAMS_KEYS = {
    "attrs",
    "button_label",
    "choices",
    "icon",
    "label",
    "name",
    "template_name",
    "theme_button_class",
}


class Toolbar(list):
    template_name = ThemeTemplate("toolbar.html")

    def __init__(self, *items, params=None, listing=None):
        if params is None:
            params = {}
        self._params = params
        self.name2item = {}
        if items:
            if isinstance(items[0], str):
                items_name = map(str.strip, items[0].split(","))
                items = []
                for name in items_name:
                    slug = re.sub(r"[_\d]+$", "", name)
                    cls = ToolbarItemMeta.get_class(slug)
                    item = cls(name)
                    item.set_listing(listing)
                    items.append(cls(name))
            elif isinstance(items[0], (list, tuple)):
                items = items[0]
            elif isinstance(items[0], GeneratorType):
                items = list(items[0])
        super().__init__(items)

    def get(self, name, default=None):
        return self.name2item.get(name, default)

    def names(self):
        return self.name2item.keys()

    def get_params(self):
        return self._params

    def bind_to_listing(self, listing):
        items = Toolbar(params=self._params)
        for item in self:
            # ensure there is one toolbarItem instance per listing
            if not item.listing:
                item = copy.copy(item)
            item.bind_to_listing(listing)
            items.append(item)
        items.name2item = {i.name: i for i in items if isinstance(i, ToolbarItem)}
        items.listing = listing
        return items

    def get_context(self):
        # aggregate all items context
        return RenderContext(*[i.get_context() for i in self])


class ToolbarItemMeta(type):
    slug2class = {}

    def __new__(mcs, name, bases, attrs):
        cls = super(ToolbarItemMeta, mcs).__new__(mcs, name, bases, attrs)
        ToolbarItemMeta.slug2class[ToolbarItemMeta.get_slug(cls)] = cls
        params_keys = cls.params_keys
        if isinstance(params_keys, str):
            params_keys = set(map(str.strip, params_keys.split(",")))
        TOOLBAR_PARAMS_KEYS.update(params_keys)
        TOOLBAR_PARAMS_KEYS.discard("")
        return cls

    @classmethod
    def get_slug(cls, mcs):
        return mcs.__name__.replace("ToolbarItem", "").lower()

    @classmethod
    def get_class(cls, slug):
        return cls.slug2class.get(slug)


class ToolbarItem(metaclass=ToolbarItemMeta):
    name = None
    params_keys = ""
    attrs = {}
    listing = None
    label = None
    icon = None
    theme_button_class = ThemeAttribute("toolbar_theme_button_class")

    _ids = count(0)

    def __init__(self, *args, **kwargs):
        self.init_args = args
        self.init_kwargs = kwargs
        self.id = next(self._ids)

    def bind_to_listing(self, listing):
        self.init(listing, *self.init_args, **self.init_kwargs)

    def set_listing(self, listing):
        self.listing = listing

    def init(self, listing, name=None, **kwargs):
        self.set_listing(listing)
        if name:
            self.name = re.sub(r"\W", "", name)
        if not self.name:
            self.name = ToolbarItemMeta.get_slug(self.__class__)
        self.set_kwargs(**kwargs)
        self.apply_template_kwargs()
        if not isinstance(self.attrs, HTMLAttributes):
            self.attrs = HTMLAttributes(self.attrs)
        self.css_class = self.__class__.__name__.lower()

    def set_kwargs(self, **kwargs):
        # If toolbar_<toolbarItemclass>_<params> exists in listing attributes :
        # take is as a default value
        for k in TOOLBAR_PARAMS_KEYS:
            listing_key = "toolbar_{}".format(k)
            if hasattr(self.listing, listing_key):
                setattr(self, k, getattr(self.listing, listing_key))
        for k, v in kwargs.items():
            if k in TOOLBAR_PARAMS_KEYS:
                setattr(self, k, v)
        # if parameters given in toolbar : apply them
        for k, v in self.listing.toolbar.get_params().get(self.name, {}).items():
            if k in TOOLBAR_PARAMS_KEYS:
                setattr(self, k, v)
        # if parameters given in listing apply them to specified toolbarItem
        for k in TOOLBAR_PARAMS_KEYS:
            key = "toolbar_{}__{}".format(self.name, k)
            v = getattr(self.listing, key, None)
            if v is not None:
                setattr(self, k, v)

    def apply_template_kwargs(self):
        kwargs = self.listing.get_toolbar_item_kwargs(self.name)
        if kwargs:
            for k, v in kwargs.items():
                if k in TOOLBAR_PARAMS_KEYS:
                    if isinstance(v, dict):
                        prev_value = getattr(self, k, None)
                        if isinstance(prev_value, dict):
                            prev_value.update(v)
                            continue
                    setattr(self, k, v)

    def get_context(self):
        return {}


class SortSelectToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_select.html")
    params_keys = "up_arrow,down_arrow,has_submit_button"
    choices = None
    up_arrow = " ↑"  # DO NOT use &#xNNNN; notation : it does not work with firefox
    down_arrow = " ↓"
    has_submit_button = False
    label = _("Sort by")
    button_label = _("OK")
    select_name = "sort"

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.choices = normalize_choices(self.choices)
        if not self.choices:
            self.choices = []
            for c in sorted(
                self.listing.columns, key=lambda x: x.get_header_value().lower()
            ):
                self.choices.append((c.name, c.get_header_value() + self.up_arrow))
                self.choices.append(
                    ("-" + c.name, c.get_header_value() + self.down_arrow)
                )

    def get_context(self):
        # get_context is called at rendering time :
        # this is the way to get the asked value for self.listing.sort
        sorting = self.listing.sort
        self.selected_choices = [
            (sort_key, label, sort_key == sorting) for sort_key, label in self.choices
        ]
        # do not need to return a context


class SortDropdownToolbarItem(SortSelectToolbarItem):
    template_name = ThemeTemplate("tbi_dropdown.html")
    label = _("Sort by...")


class VariationsToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_variations.html")
    params_keys = [
        "show_labels",
        "show_icons",
        "labels",
        "icons",
        "show_tooltip",
        "menu_css_class",
    ]
    labels = ""
    icons = ""
    show_labels = False
    show_icons = True
    show_tooltip = True
    buttons = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        if not hasattr(self.listing, "variations_classes"):
            raise InvalidToolbarItem(
                "Variation toolbar item is only for variations listings "
                "(subclass of ListingVariations)"
            )
        nb_variations = len(self.listing.variations_classes)
        labels = normalize_list(self.labels, force_length=nb_variations)
        icons = normalize_list(self.icons, force_length=nb_variations)
        urls = [
            self.listing.get_url(variation=i, without="gb_cols,gb_annotate_cols")
            for i, cls in enumerate(self.listing.variations_classes)
        ]
        self.buttons = list(
            zip(
                urls,
                labels,
                icons,
            )
        )
        variation = max(0, min(self.listing.variation, nb_variations - 1))
        self.icon = self.icons[variation]
        self.label = self.labels[variation]


class VariationsDropdownToolbarItem(VariationsToolbarItem):
    template_name = ThemeTemplate("tbi_variations_dropdown.html")
    params_keys = ["show_labels", "show_icons", "labels", "icons"]
    labels = ""
    icons = ""
    show_labels = True
    show_icons = True


class PerPageSelectToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_select.html")
    params_keys = "choices"
    choices = "10,25,50,100,-1:All"
    has_submit_button = False
    label = _("Per page")
    button_label = _("OK")
    select_name = "per_page"

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.choices = normalize_choices(self.choices, int_keys=True)

    def get_context(self):
        # get_context is called at rendering time :
        # this is the way to get the asked value for self.listing.per_page
        # if per_pages is not in choices : add that choice dynamically
        per_pages = [c[0] for c in self.choices]
        per_page = self.listing.per_page
        if per_pages and per_page not in per_pages:
            for i, p in enumerate(per_pages):
                if per_page < p or p == -1:
                    self.choices.insert(i, [per_page, per_page])
                    break
            else:
                self.choices.append([per_page, per_page])

        self.selected_choices = [
            (pp, label, per_page == pp) for pp, label in self.choices
        ]
        # do not need to return a context


class PerPageDropdownToolbarItem(PerPageSelectToolbarItem):
    template_name = ThemeTemplate("tbi_dropdown.html")
    label = _("Per page...")


class ExportSelectToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_export.html")
    choices = [("", _("Choose..."))] + EXPORT_FORMATS
    has_submit_button = True
    label = _("Export to...")
    button_label = _("OK")
    select_name = "export"

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)
        self.choices = normalize_choices(self.choices)
        self.selected_choices = [(key, label, False) for key, label in self.choices]


class ExportDropdownToolbarItem(ExportSelectToolbarItem):
    template_name = ThemeTemplate("tbi_export_dropdown.html")
    label = _("Export to...")
    choices = EXPORT_FORMATS


class UpdateToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_update.html")
    label = _("Update")


class SelectAllToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_select_all.html")
    label = _("Select all")


class UnselectAllToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_unselect_all.html")
    label = _("Unselect all")


class InvertSelectionToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_invert_selection.html")
    label = _("Invert selection")


class LinkButtonToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_link_button.html")
    params_keys = ["url", "url_name", "querystring", "action"]
    label = _("** link button **")
    url = "#"
    url_name = None
    querystring = ""
    action = None

    def get_context(self):
        if callable(self.url):
            self.url = self.url(self)
        elif self.url_name:
            self.url = reverse(self.url_name)
        if self.querystring:
            if self.querystring.startswith("?"):
                self.querystring = self.querystring[1:]
            self.url += "?" + self.querystring


class GroupByToolbarItem(ToolbarItem):
    template_name = ThemeTemplate("tbi_group_by.html")
    label = _("Group by...")

    def get_context(self):
        self.listing.need_media_for("dual_listbox")
        return dict(gb_cols=self.listing.gb_cols)



#
# Created : 2018-02-09
#
# @author: Eric Lapouyade
#
import base64
import collections
import os
import re
import types
from urllib.parse import quote_plus

from django.core.exceptions import FieldDoesNotExist
from django.core.serializers.json import Serializer
from django.db import models
from django.db.models import F, Model
from django.db.models.query import QuerySet
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from . import FILTER_QUERYSTRING_PREFIX
from .exceptions import *
from .utils import to_js_timestamp

__all__ = [
    "RecordManager",
    "Record",
    "cache_in_record",
    "object_serializer",
    "FORM_LABEL_METHOD_NAME",
]


FORM_LABEL_METHOD_NAME = "get_form_label"


def cache_in_record(value_func):
    def cached_value_func(self, rec):
        cached, value = rec.get_cached_cell_value(self)
        if not cached:
            value = value_func(self, rec)
            rec.set_cached_cell_value(self, value)
        return value

    return cached_value_func


class SequenceItem:
    def __init__(self, pk, obj):
        self.pk = pk
        self.obj = obj


class ObjectSerializer(Serializer):
    def serialize(self, *args, **kwargs):
        self.additional_data = kwargs.pop("additional_data", None)
        self.form_data = kwargs.pop("form_data", None)
        return super().serialize(*args, **kwargs)

    def start_serialization(self):
        self._init_options()

    def end_serialization(self):
        pass

    def get_dump_object(self, obj):
        data = {"fields": self._current}
        data["fields"][obj._meta.pk.attname] = obj.pk
        if self.additional_data:
            data["data"] = self.additional_data
        if self.form_data:
            data["formfields"] = self.form_data
        return data


object_serializer = ObjectSerializer()


# Records are always bound to a listing instance
class RecordManager:
    def __init__(self, listing):
        self.listing = listing
        self._records = None

    def get_all(self):
        # used only for sequences (and short sequences please !)
        lsg = self.listing
        if isinstance(lsg.data, QuerySet):
            raise InvalidListing(
                _(
                    "Getting all records is only supported on "
                    "SMALL sequence data, not on Django querysets"
                )
            )
        if not hasattr(lsg, "_all_records"):
            lsg._all_records = [Record(lsg, rec, i) for i, rec in enumerate(lsg.data)]
        return lsg._all_records

    def get_first_obj(self):
        if not hasattr(self, "_first_obj"):
            qs = self.get_objs_from_queryset()
            self._first_obj = qs.first()
        return self._first_obj

    def get_last_obj(self):
        if not hasattr(self, "_last_obj"):
            qs = self.get_objs_from_queryset()
            self._last_obj = qs.last()
        return self._last_obj

    def clear_first_last_cache(self):
        if hasattr(self, "_first_obj"):
            delattr(self, "_first_obj")
        if hasattr(self, "_last_obj"):
            delattr(self, "_last_obj")

    def get_obj(self, **kwargs):
        lsg = self.listing
        if isinstance(lsg.data, QuerySet):
            qs = self.get_objs_from_queryset()
            try:
                return qs.get(**kwargs)
            except qs.DoesNotExist:
                return None
        else:
            # actually work only with 'id' or 'pk'
            pk = kwargs.get("pk")
            if pk is None:
                pk = kwargs.get("id")
            if pk is None:
                return None
            return self.get_objs_from_sequence()[pk].obj

    def get_unfiltered_count(self):
        data = self.listing.data
        if isinstance(data, QuerySet):
            return data.count()
        else:
            return len(data)

    def get_rec(self, **kwargs):
        obj = self.get_obj(**kwargs)
        if obj is None:
            return None
        return Record(self.listing, obj)

    def move_obj_order(self, obj, field, delta):
        qs = self.get_objs_from_queryset()
        obj_order = getattr(obj, field)
        if not isinstance(obj_order, int):
            raise ListingException(
                _(
                    'field "{field}" must be an integer, please choose the '
                    "correct field that will be used to change object ordering"
                ).format(field=field)
            )
        new_order = obj_order + delta
        if delta < 0:
            qs = qs.filter(**{f"{field}__gte": new_order, f"{field}__lt": obj_order})
            qs.update(**{field: F(field) + 1})
            setattr(obj, field, new_order)
            obj.save()
            self.clear_first_last_cache()
        elif delta > 0:
            qs = qs.filter(**{f"{field}__gt": obj_order, f"{field}__lte": new_order})
            qs.update(**{field: F(field) - 1})
            setattr(obj, field, new_order)
            obj.save()
            self.clear_first_last_cache()

    def compute_current_page_records(self):
        lsg = self.listing
        if not isinstance(lsg.data, (collections.abc.Sequence, QuerySet)):
            raise InvalidData(_("Listing data must be a sequence or a QuerySet."))
        if isinstance(lsg.data, QuerySet):
            objs = self.get_objs_from_queryset()
        else:
            objs = self.get_objs_from_sequence()

        per_page = int(lsg.per_page)
        if per_page <= 0 or per_page > lsg.per_page_max:
            per_page = lsg.per_page_max
        lsg.paginator = lsg.paginator_class(
            lsg, objs, per_page, lsg.orphans, lsg.allow_empty_first_page
        )
        lsg.current_page = lsg.paginator.get_page(lsg.page or 1)
        if lsg.editable and lsg.editing:
            self.bind_formset()

    def group_by_model_field_map(self, records, col, model, field):
        try:
            if "__" in field:
                # if field is "aa__bb__cc"
                # then left = aa and right == bb__cc
                left, right = re.match(r"^(.*?)(?:__)(.*?)$", field).groups()
                model_field = model._meta.get_field(left)
                if model_field.__class__.__name__ == "ForeignKey":
                    # recurse to the right part of the field
                    self.group_by_model_field_map(
                        records, col, model_field.related_model, right
                    )
            else:
                model_field = model._meta.get_field(field)
                if model_field.__class__.__name__ == "ForeignKey":
                    # As it is a group by queryset, ForeignKey's does not give object
                    # But only their PK's, so translating that into objects...
                    rec_key = col.data_key
                    pks = [rec.get(rec_key) for rec in records]
                    objs = model_field.related_model.objects.filter(pk__in=pks)
                    pk2obj = {obj.pk: obj for obj in objs}
                    for rec in records:
                        rec.set(col.data_key, pk2obj.get(rec.get(rec_key)))
        except FieldDoesNotExist:
            pass

    def group_by_foreignkey_object_map(self, records):
        lsg = self.listing
        if lsg.gb_cols:
            # Remap col where data_key is not part of the main model
            for col in lsg.columns:
                self.group_by_model_field_map(records, col, lsg.model, col.data_key)

    def current_page(self):
        if not self._records:
            lsg = self.listing
            self._records = [
                # lsg.current_page is set in RecordManager.compute_current_page_records()
                Record(lsg, obj, i)
                for i, obj in enumerate(lsg.current_page)
            ]
            self.group_by_foreignkey_object_map(self._records)
            lsg.update_page_records(self._records)
        return self._records

    def bind_formset(self):
        if self.listing.editing_row_pk:
            formset = self.listing.get_formset()
            for rec in self.current_page():
                if rec.pk == self.listing.editing_row_pk:
                    rec.set_form(formset[0])
        else:
            for rec, form in zip(self.current_page(), self.listing.get_formset()):
                rec.set_form(form)

    def export(self):
        lsg = self.listing
        export_data = self.get_export_data()
        export_data = self.order_data(export_data)
        if lsg.gb_cols:
            records = [Record(lsg, obj, i) for i, obj in enumerate(export_data)]
            self.group_by_foreignkey_object_map(records)
            for rec in records:
                yield rec
        else:
            for i, obj in enumerate(export_data):
                yield Record(lsg, obj, i)

    def get_export_data(self):
        data = self.listing.data
        if isinstance(data, QuerySet):
            export_data = self.filter_queryset(data)
        else:
            export_data = self.filter_sequence(data)
        return export_data

    def export_count(self):
        export_data = self.get_export_data()
        if isinstance(export_data, QuerySet):
            return export_data.count()
        else:
            return len(export_data)

    def filter_queryset(self, qs):
        if self.listing.filters:
            qs = self.listing.filters.filter_queryset(qs)
        return qs

    def get_filtered_queryset(self):
        qs = self.listing.data
        qs = self.filter_queryset(qs)
        return qs

    def order_data(self, data):
        if isinstance(data, QuerySet):
            data = self.order_queryset(data)
        else:
            data = self.order_sequence(data)
        return data

    def get_order_by(self):
        lsg = self.listing
        if lsg.force_order_by:
            if isinstance(lsg.force_order_by, str):
                order_by = (lsg.force_order_by,)
            else:
                order_by = lsg.force_order_by
        else:
            order_by = []
            if lsg.sort:
                for col_name in lsg.columns_sort_list:
                    order_prefix = "" if lsg.columns_sort_ascending[col_name] else "-"
                    col = lsg.columns.get(col_name)
                    if col:
                        if isinstance(col.sort_key, (tuple, list)):
                            for sort_key in col.sort_key:
                                order_by.append(order_prefix + sort_key)
                        else:
                            order_by.append(order_prefix + col.sort_key)
            # add a default sorting to avoid a Django 'UnorderedObjectListWarning'
            if not order_by:
                if lsg.gb_cols:
                    # if "group by" feature activated
                    order_by = lsg.gb_queryset_fields
                else:
                    order_by = ["pk"]
        return order_by

    def order_queryset(self, qs):
        order_by = self.get_order_by()
        return qs.order_by(*order_by)

    def order_sequence(self, data):
        # only manage one level sort
        order_by = self.get_order_by()
        if order_by:
            sort_key = order_by[0]
            sort_desc = sort_key[0] == "-"
            if sort_desc:
                sort_key = sort_key[1:]
            data = sorted(data, key=lambda x: x[sort_key], reverse=sort_desc)
        return data

    def get_objs_from_queryset(self):
        if not hasattr(self, "_queryset_objs"):
            qs = self.get_filtered_queryset()
            qs = self.order_queryset(qs)
            self._queryset_objs = qs
        return self._queryset_objs

    def filter_sequence(self, seq):
        if self.listing.filters:
            for filtr in self.listing.filters:
                seq = filtr.filter_sequence(seq)
        return seq

    def get_objs_from_sequence(self):
        lsg = self.listing
        seq = self.listing.data
        seq = self.filter_sequence(seq)
        if lsg.sort:
            for col_name in reversed(lsg.columns_sort_list):
                col = lsg.columns.get(col_name)
                if col:
                    sort_key = col.sort_key
                    if callable(sort_key):
                        key_func = sort_key
                    else:

                        def key_func(rec):
                            return rec[sort_key]

                    seq = sorted(
                        seq,
                        key=key_func,
                        reverse=not lsg.columns_sort_ascending[col_name],
                    )
        return [SequenceItem(i, obj) for i, obj in enumerate(seq)]


class Record:
    def __init__(self, listing, obj, index=0, form=None):
        self._listing = listing
        if isinstance(obj, SequenceItem):
            self._obj = obj.obj
            self.pk = obj.pk
        else:
            self._obj = obj
        self._index = index
        self._cell_values = {}
        self._form = form
        self._selected = False
        self._is_qs_first = False
        self._is_qs_last = False
        self._qs_record_index = 0
        if isinstance(self._obj, dict):
            self._format_ctx = self._obj
        elif isinstance(self._obj, (tuple, list)):
            self._format_ctx = {
                self._listing.columns[i].name: val for i, val in enumerate(self._obj)
            }
        else:
            self._format_ctx = vars(self._obj)
        cp = listing.current_page
        if cp:
            self._qs_record_index = cp.start_index() + index  # 1-based index
            if self._qs_record_index == 1:
                self._is_qs_first = True
            if self._qs_record_index == cp.paginator.count:
                self._is_qs_last = True

    def get_object(self):
        return self._obj

    def get_in_obj(self, key, default=None):
        return self._obj.get(key, default)

    def get_form_serialized_cols(self, obj, cols):
        data = {}
        for col in cols:
            if col.form_field_serialize_func:
                value = col.form_field_serialize_func(col, obj)
                if value is not None:
                    data[col.name] = value
                continue
            data_key = col.data_key
            from django_listing import MultiAutoCompleteColumn

            if isinstance(col, MultiAutoCompleteColumn):
                m2m_qs = getattr(obj, data_key).all()
                m2m_pks_labels = []
                for m2m_obj in m2m_qs:
                    form_label_func = getattr(m2m_obj, FORM_LABEL_METHOD_NAME, None)
                    if form_label_func is None:
                        form_label_func = lambda: str(m2m_obj)
                    m2m_pks_labels.append((m2m_obj.pk, form_label_func()))
                data[col.name] = m2m_pks_labels
                continue

            final_object = None
            if "__" in data_key:
                attr, foreign_attr = data_key.split("__", maxsplit=1)
                foreign_object = getattr(obj, attr, None)
                final_object = getattr(foreign_object, foreign_attr, None)
            else:
                final_object = getattr(obj, data_key, None)

            from django_listing import AutoCompleteColumn

            if (
                isinstance(final_object, int)
                # use class name to avoid cyclic import
                and isinstance(col, AutoCompleteColumn)
                and hasattr(col, "queryset")
            ):
                final_object = col.queryset.filter(pk=final_object).first()
            if final_object is None:
                continue
            if isinstance(final_object, Model):
                form_label_func = getattr(final_object, FORM_LABEL_METHOD_NAME, None)
                if form_label_func is None:
                    form_label_func = lambda: str(final_object)
                data[col.name] = (final_object.pk, form_label_func())
            else:
                # final_object is a value, not a model object
                data[col.name] = final_object
        return data

    def get_serialized_object(self, **kwargs):
        method = getattr(self._obj, "get_serialized_additional_data", None)
        additional_data = method() if method else {}
        form_data = {}
        if self._listing.form_serialize_cols:
            func = self._listing.form_serialize_cols_func
            if func is None:
                func = self.get_form_serialized_cols
            elif isinstance(func, str):
                func = getattr(self._obj.__class__, func)
            form_data.update(func(self._obj, self._listing.form_serialize_cols))
        if self._listing.form_no_autofill_cols:
            additional_data["no_autofill"] = self._listing.form_no_autofill_cols
        method = getattr(self._obj, "get_serialized_form_data", None)
        if method:
            form_data.update(method())
        serialized_obj = object_serializer.serialize(
            [self._obj],
            form_data=form_data,
            additional_data=additional_data,
            separators=(",", ":"),
            **kwargs,
        )
        return base64.b64encode(serialized_obj.encode()).decode()

    def is_first_qs_record(self):
        return self._is_qs_first

    def is_last_qs_record(self):
        return self._is_qs_last

    def is_selected(self):
        return self._selected

    def set_selected(self):
        self._selected = True

    def get_form(self):
        return self._form

    def set_form(self, form):
        self._form = form

    def get_cached_cell_value(self, col):
        if col in self._cell_values:
            return True, self._cell_values[col]
        return False, None

    def set_cached_cell_value(self, col, value):
        self._cell_values[col] = value

    def get_format_ctx(self):
        return self._format_ctx

    def format_str(self, str_to_format, extra_context=None):
        context = {**self._format_ctx}
        obj = self.get_object()
        context.update(rec_object=obj)
        if hasattr(obj, "get_absolute_url"):
            context.update(rec_object_url=obj.get_absolute_url())
        return mark_safe(str_to_format.format(**context))

    def get_listing(self):
        return self._listing

    def get_index(self):
        return self._index

    def do_filter_object(self, obj, filter_name, params):
        if filter_name == "urlencode" and isinstance(obj, str):
            obj = quote_plus(obj)
        elif filter_name == "replace":
            obj = obj.replace(*params)
        elif filter_name == "sub":
            obj = re.sub(*params, obj)
        elif filter_name == "basename":
            obj = os.path.basename(obj)
        elif filter_name == "js_timestamp":
            obj = to_js_timestamp(obj)
        else:
            func = getattr(obj, filter_name, None)
            if func is not None:
                obj = func()
        return obj

    def get_href(self):
        if hasattr(self._obj, "get_absolute_url"):
            return self._obj.get_absolute_url()

    def get_url(self, filters=None, **kwargs):
        if isinstance(filters, dict):
            for filter_name, rec_key in filters.items():
                fname = filter_name + self._listing.suffix
                val = self.get(rec_key)
                if isinstance(val, Model):
                    val = val.pk
                kwargs[f"{FILTER_QUERYSTRING_PREFIX}{fname}"] = val
        return self._listing.get_url(**kwargs)

    def get(self, key, default=None):
        obj = self._obj
        # case obj is a dict, try without splitting key (case group by)
        if isinstance(obj, dict) and key in obj:
            return obj.get(key)
        try:
            if isinstance(key, int):
                obj = obj[key]
            else:
                key, *filter_names = key.split("|", 1)
                for subkey in key.split("__"):
                    if "|" in subkey:
                        subkey, *filter_names = subkey.split("|", 1)
                    if subkey.isdigit():
                        obj = obj[int(subkey)]
                    else:
                        if isinstance(obj, dict):
                            obj = obj[subkey]
                        else:
                            obj = getattr(obj, subkey)
                    if isinstance(obj, types.MethodType):
                        obj = obj()
                if obj is not None and filter_names:
                    filter_names = filter_names[0]
                    filter_name, *params = re.split(r"(?<!\\):", filter_names)
                    params = tuple(map(lambda s: re.sub(r"\\:", ":", s), params))
                    obj = self.do_filter_object(obj, filter_name, params)
        except (
            IndexError,
            AttributeError,
            TypeError,
            KeyError,
            models.ObjectDoesNotExist,
        ):
            return default
        return obj

    def _digest(self):
        out = ""
        for c in self._listing.columns:
            out += "<b>{} :</b> {}<br>\n".format(c.name, c.get_cell_value(self))
        return out

    def __str__(self):
        return mark_safe(self._digest())

    def __getattr__(self, item):
        return self.get(item)

    def __getitem__(self, item):
        return self.get(item)

    def __delitem__(self, key):
        del self._obj[key]

    def set(self, item, value):
        if isinstance(self._obj, dict):
            self._obj[item] = value
        else:
            setattr(self._obj, item, value)



#
# Created : 2018-24-04
#
# @author: Eric Lapouyade
#

from django.db.models import Avg, Max, Min, Sum
from django.db.models.query import QuerySet
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from .exceptions import InvalidAggregation

__all__ = [
    "Aggregation",
    "AggregationMeta",
    "AvgAggregation",
    "MaxAggregation",
    "MinAggregation",
    "SumAggregation",
]


class AggregationMeta(type):
    slug2class = {}

    def __new__(mcs, name, bases, attrs):
        cls = super(AggregationMeta, mcs).__new__(mcs, name, bases, attrs)
        AggregationMeta.slug2class[AggregationMeta.get_slug(cls)] = cls
        return cls

    @classmethod
    def get_slug(cls, mcs):
        return mcs.__name__.replace("Aggregation", "").lower()

    @classmethod
    def get_aggregation_slugs(cls):
        aggs = list(filter(None, cls.slug2class.keys()))
        return aggs + ["global_" + agg for agg in aggs]

    @classmethod
    def get_instance(cls, slug, column):
        global_aggregation = False
        if slug.startswith("global_"):
            global_aggregation = True
            slug = slug[7:]
        agg_cls = cls.slug2class.get(slug)
        if agg_cls is None:
            raise InvalidAggregation(
                gettext('Unknown "{}" aggregation, choose one of these : ' "{}").format(
                    slug, ",".join(cls.get_aggregation_slugs())
                )
            )
        return agg_cls(column, global_aggregation=global_aggregation)


class Aggregation(metaclass=AggregationMeta):
    footer_tpl = None
    value_tpl = "{value}"
    params_keys = None

    def __init__(self, column, global_aggregation=False):
        self.values = []
        self.column = column
        self.global_aggregation = global_aggregation
        if self.params_keys:
            params_to_check = self.params_keys
            if isinstance(params_to_check, str):
                params_to_check = params_to_check.split(",")
            for p in params_to_check:
                if not hasattr(column, p):
                    setattr(column, p, getattr(self, p))

    def get_numeric_values(self):
        return [v for v in self.values if isinstance(v, (int, float))]

    def aggregate(self, rec):
        if not self.global_aggregation:
            value = self.column.get_cell_value(rec)
            self.values.append(value)

    def get_footer_attrs(self, ctx, value):
        attrs = self.column.get_footer_attrs(ctx, value)
        attrs.add("class", "agg-{}".format(AggregationMeta.get_slug(self.__class__)))
        return attrs

    def get_footer_context(self, value):
        return self.column.get_footer_context(value)

    def get_footer_value_tpl(self, ctx, value):
        col = self.column
        if hasattr(col, "get_footer_aggregation_value_tpl"):
            return col.get_footer_aggregation_value_tpl(ctx, value)
        elif col.footer_value_tpl is not None:
            return col.footer_value_tpl
        elif self.value_tpl is not None:
            return self.value_tpl
        return "{value}"

    def get_footer_template(self, ctx, value):
        col = self.column
        if hasattr(col, "get_footer_aggregation_template"):
            return col.get_footer_aggregation_template(ctx, value)
        elif col.footer_tpl is not None:
            return col.footer_tpl
        elif self.footer_tpl is not None:
            return self.footer_tpl
        return "<td{attrs}>%s</td>" % self.get_footer_value_tpl(ctx, value)

    def render_footer(self):
        if self.values or self.global_aggregation:
            value = self.get_aggregated_value()
            ctx = self.get_footer_context(value)
            ctx.attrs = self.get_footer_attrs(ctx, value)
            tpl = self.get_footer_template(ctx, value)
            try:
                return tpl.format(**ctx)
            except (ValueError, AttributeError, IndexError) as e:
                return '<td class="render-error">{}</td>'.format(e)
        else:
            return self.column.default_footer_value

    def get_aggregated_value_from_sequence(self, seq):
        records = self.column.listing.records.get_all()
        self.values = [self.column.get_cell_value(rec) for rec in records]
        return self.get_aggregated_value_from_current_page()

    def get_aggregated_value(self):
        listing = self.column.listing
        if self.global_aggregation:
            data = listing.data
            if isinstance(data, QuerySet):
                data = listing.records.get_filtered_queryset()
                return self.get_aggregated_value_from_queryset(data)
            else:
                return self.get_aggregated_value_from_sequence(data)
        else:
            return self.get_aggregated_value_from_current_page()


class SumAggregation(Aggregation):
    value_tpl = _("Total :<br>{value}")

    def get_aggregated_value_from_current_page(self):
        return sum(self.get_numeric_values())

    def get_aggregated_value_from_queryset(self, qs):
        return qs.aggregate(val=Sum(self.column.data_key)).get("val")


class MinAggregation(Aggregation):
    value_tpl = _("Min :<br>{value}")

    def get_aggregated_value_from_current_page(self):
        return min(self.get_numeric_values())

    def get_aggregated_value_from_queryset(self, qs):
        return qs.aggregate(val=Min(self.column.data_key)).get("val")


class MaxAggregation(Aggregation):
    value_tpl = _("Max :<br>{value}")

    def get_aggregated_value_from_current_page(self):
        return max(self.get_numeric_values())

    def get_aggregated_value_from_queryset(self, qs):
        return qs.aggregate(val=Max(self.column.data_key)).get("val")


class MinMaxAggregation(Aggregation):
    value_tpl = _("Min : {min_val}<br>Max : {max_val}")

    def get_aggregated_value_from_current_page(self):
        return dict(
            min_val=min(self.get_numeric_values()),
            max_val=max(self.get_numeric_values()),
        )

    def get_aggregated_value_from_queryset(self, qs):
        return qs.aggregate(
            min_val=Min(self.column.data_key), max_val=Max(self.column.data_key)
        )


class MinMaxAvgAggregation(Aggregation):
    value_tpl = _(
        "Min : {min_val}<br>Max : {max_val}<br>"
        "Avg : {avg_val:.{col.footer_precision}f}"
    )
    params_keys = "footer_precision"
    footer_precision = 2

    def get_aggregated_value_from_current_page(self):
        num_values = self.get_numeric_values()
        return dict(
            min_val=min(num_values),
            max_val=max(num_values),
            avg_val=sum(num_values) / len(num_values),
        )

    def get_aggregated_value_from_queryset(self, qs):
        return qs.aggregate(
            min_val=Min(self.column.data_key),
            max_val=Max(self.column.data_key),
            avg_val=Avg(self.column.data_key),
        )


class AvgAggregation(Aggregation):
    value_tpl = "Average :<br>{value:.{col.footer_precision}f}"
    params_keys = "footer_precision"
    footer_precision = 2

    def get_aggregated_value_from_current_page(self):
        num_values = self.get_numeric_values()
        return sum(num_values) / len(num_values)

    def get_aggregated_value_from_queryset(self, qs):
        return qs.aggregate(val=Avg(self.column.data_key)).get("val")



#
# Created : 2018-02-09
#
# @author: Eric Lapouyade
#

import collections
import logging
import re
from datetime import datetime
from urllib.parse import urlsplit, urlunsplit, quote

import tablib
from django import forms
from django.conf import settings
from django.contrib import messages
from django.db.models import Model, Count, Min, Max, Sum, Avg
from django.db.models.query import QuerySet
from django.http import QueryDict
from django.middleware.csrf import get_token as get_csrf_token
from django.template import loader
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy, pgettext_lazy
from django.utils import timezone

from django_listing import (
    EXPORT_FORMATS,
    EXPORT_FORMATS_KEEP_ORIGINAL_TYPE,
    EXPORT_FORMATS_USE_COL_NAME,
    EXPORT_EXCEL_SANITIZE_RE,
    FILTER_QUERYSTRING_PREFIX,
)

from .columns import (
    COLUMNS_PARAMS_KEYS,
    ModelColumns,
    SelectionColumn,
    SequenceColumns,
    Columns,
    IntegerColumn,
    Column,
    GroupByFilterColumn,
    FloatColumn,
)
from .context import RenderContext
from .exceptions import *
from .filters import FILTERS_PARAMS_KEYS, Filters
from .html_attributes import HTMLAttributes
from .attached_form import ATTACHED_FORM_PARAMS_KEYS, ListingBaseForm
from .paginators import PAGINATOR_PARAMS_KEYS, Paginator
from .record import RecordManager
from .theme_config import ThemeAttribute, ThemeTemplate
from .toolbar import TOOLBAR_PARAMS_KEYS, Toolbar
from .utils import init_dicts_from_class, validate_values_names

__all__ = ["ListingVariations", "Listing", "DivListing", "logger"]

logger = logging.getLogger("django_listing")

LISTING_OFFSET_MAX = 100_000
LISTING_ROWS_PER_PAGE = 20
LISTING_ROWS_PER_PAGE_MAX = 500
LISTING_SUFFIX_REQUEST_DATA_FIELD = "_listing_suffix_data"
LISTING_SUFFIX_PATTERN = "-{}"
LISTING_SELECTION_INPUT_NAME_KEY = "selected_rows"
LISTING_SELECTION_CHECKBOX_NAME = "selection_checkbox"
LISTING_SELECTION_HOVER_CSS_CLASS = "hover"
LISTING_SELECTOR_CSS_CLASS = "row-selector"

# only these parameters are allowed in the query string.
LISTING_QUERY_STRING_KEYS = {
    "action_button",
    "editing",
    "editing_columns",
    "editing_row_pk",
    "export",
    "page",
    "per_page",
    "select_columns",
    "selecting",
    "sort",
    "theme",
    "variation",
    "gb_cols",
    "gb_annotate_cols",
}
LISTING_QUERY_STRING_INT_KEYS = {"page", "per_page", "variation", "editing_row_pk"}
LISTING_NOT_PERSISTENT_QUERY_STRING_KEYS = set()

# Only these attributes can be modified outside the class (in django templates,
# views, listings instances ...)
LISTING_PARAMS_KEYS = {
    "accept_ajax",
    "action_button_cancel_label",
    "action_button_edit_label",
    "action_button_update_label",
    "action_button_upload_label",
    "action_footer_template_name",
    "action_header_template_name",
    "ajax_part",
    "ajax_request",
    "allow_empty_first_page",
    "anchor_hash",
    "attrs",
    "chart_options_template",
    "columns_headers",
    "confirm_msg_for_delete",
    "confirm_msg_nb_items_for_delete",
    "confirm_msg_for_update",
    "confirm_msg_nb_items_for_update",
    "container_attrs",
    "data",
    "datetimepicker_date_format",
    "datetimepicker_datetime_format",
    "datetimepicker_time_format",
    "div_row_template_name",
    "div_rows",
    "div_template_name",
    "edit_on_demand",
    "editable",
    "editable_columns",
    "editing",
    "editing_columns",
    "editing_hidden_columns",
    "editing_row_pk",
    "empty_listing_template_name",
    "empty_table_msg",
    "exclude_columns",
    "export",
    "export_columns",
    "export_max_rows",
    "filters",
    "footer_snippet",
    "footer_template_name",
    "form",  # usually insert/edit form (NOT filter form)
    "format_numbers",
    "attached_form_autofill",  # insert/edit form is automatically filled when a listing row is selected
    "gb_annotate_cols",  # group_by annotation columns
    "gb_cols",  # group_by columns
    "gb_count_layout",  # can be "begin", "end" or None
    "global_context",
    "gb_template_name",
    "has_footer",
    "has_footer_action_buttons",
    "has_gb_filter_button",
    "has_header",
    "has_nb_unfiltered_rows",
    "has_paginator",
    "has_upload",
    "id",
    "link_object_columns",
    "listing_template_name",
    "name",
    "no_permission_required_for_actions",
    "offset_max",
    "onready_snippet",
    "orphans",
    "page",
    "paginator_class",
    "per_page",
    "per_page_max",
    "primary_key",
    "processed_flash",
    "processed_pks",
    "record_label",
    "record_label_plural",
    "request_media_for",
    "row_attrs",
    "row_form_base_class",
    "save_to_database",
    "select_columns",
    "selectable",
    "selecting",
    "selection_initial",
    "selection_key",
    "selection_menu_id",
    "selection_mode",
    "selection_multiple",
    "selection_multiple_ctrl",
    "selection_overlay_template_name",
    "selection_position",
    "small_device_header_style",
    "sort",
    "sortable",
    "suffix",
    "theme",
    "theme_action_button_cancel_icon",
    "theme_action_button_class",
    "theme_action_button_edit_icon",
    "theme_action_button_update_icon",
    "theme_action_button_upload_icon",
    "theme_button_active_class",
    "theme_button_disabled_class",
    "theme_div_row_container_class",
    "theme_localized_small_device_styles_width",
    "theme_sort_asc_class",
    "theme_sort_asc_icon",
    "theme_sort_desc_class",
    "theme_sort_desc_icon",
    "theme_sort_none_icon",
    "theme_sortable_class",
    "theme_sorted_class",
    "theme_spinner_icon",
    "theme_row_class",
    "theme_table_header_class",
    "toolbar",
    "toolbar_placement",
    "unsortable",
    "use_datetimepicker",
    "variation",
    "variations",
}

attached_formset_prefix = "listing"

LISTING_ANNOTATIONS = {
    "min": (pgettext_lazy("abbreviation", "Min"), Min),
    "max": (pgettext_lazy("abbreviation", "Max"), Max),
    "sum": (pgettext_lazy("abbreviation", "Sum"), Sum),
    "avg": (pgettext_lazy("abbreviation", "Avg"), Avg),
}


class ListingBase:
    PERMISSION_ACTIONS = dict(
        insert="add",
        duplicate="add",
        update="change",
        update_all="change",
        delete_all="delete",
    )

    def __init__(self, *args, **kwargs):
        self.stored_params = {}
        self.stored_columns_params = {}
        self.stored_toolbar_items_params = {}
        self.stored_filters_params = {}
        self.stored_form_params = {}

    def transfert_params_from(self, listing):
        self.stored_params = listing.stored_params
        self.stored_columns_params = listing.stored_columns_params
        self.stored_toolbar_items_params = listing.stored_toolbar_items_params
        self.stored_filters_params = listing.stored_filters_params
        self.stored_form_params = listing.stored_form_params

    def set_attr(self, attr, value):
        setattr(self, attr, value)

    def store_kwargs(self, **kwargs):
        self.stored_params.update(kwargs)

    def store_html_attr(self, listing_attr, html_attr, value):
        html_params = self.stored_params.get(listing_attr)
        if html_params is None:
            self.stored_params[listing_attr] = HTMLAttributes({html_attr: value})
        else:
            html_params.add(html_attr, value)

    def store_column_kwargs(self, name, **kwargs):
        self.stored_columns_params.setdefault(name, {}).update(kwargs)

    def store_column_html_attr(self, name, col_attr, html_attr, value):
        col_params = self.stored_columns_params.get(name)
        if col_params is None:
            param = {col_attr: HTMLAttributes({html_attr: value})}
            self.store_column_kwargs(name, **param)
        else:
            html_params = col_params.setdefault(html_attr, HTMLAttributes())
            html_params.add(html_attr, value)

    def store_toolbar_item_html_attr(self, name, tbi_attr, html_attr, value):
        tbi_params = self.stored_toolbar_items_params.get(name)
        if tbi_params is None:
            param = {tbi_attr: HTMLAttributes({html_attr: value})}
            self.store_toolbar_item_kwargs(name, **param)
        else:
            html_params = tbi_params.setdefault(html_attr, HTMLAttributes())
            html_params.add(html_attr, value)

    def get_column_kwargs(self, name):
        return self.stored_columns_params.get(name)

    def store_toolbar_item_kwargs(self, name, **kwargs):
        self.stored_toolbar_items_params.setdefault(name, {}).update(kwargs)

    def get_toolbar_item_kwargs(self, name):
        return self.stored_toolbar_items_params.get(name)

    def store_filter_kwargs(self, name, **kwargs):
        self.stored_filters_params.setdefault(name, {}).update(kwargs)

    def store_filter_html_attr(self, name, filter_attr, html_attr, value):
        filter_params = self.stored_filters_params.get(name)
        if filter_params is None:
            param = {filter_attr: HTMLAttributes({html_attr: value})}
            self.store_filter_kwargs(name, **param)
        else:
            html_params = filter_params.setdefault(html_attr, HTMLAttributes())
            html_params.add(html_attr, value)

    def get_filter_kwargs(self, name):
        return self.stored_filters_params.get(name)

    def store_form_kwargs(self, name, **kwargs):
        self.stored_form_params.setdefault(name, {}).update(kwargs)

    def store_form_html_attr(self, name, form_attr, html_attr, value):
        form_params = self.stored_form_params.get(name)
        if form_params is None:
            param = {form_attr: HTMLAttributes({html_attr: value})}
            self.store_form_kwargs(name, **param)
        else:
            html_params = form_params.setdefault(html_attr, HTMLAttributes())
            html_params.add(html_attr, value)

    def get_form_kwargs(self, name):
        return self.stored_form_params.get(name)


class ListingVariations(ListingBase):
    variations_classes = ()
    variation_default = 0
    data = None
    listing = None
    id = None
    suffix = None

    def __init__(self, data=None, **kwargs):
        super().__init__(data, **kwargs)
        self._view = None
        if data is not None:
            self.init(data, **kwargs)
        self.init_kwargs = kwargs

    @property
    def model(self):
        if self.listing:
            return self.listing.get_model()
        return None

    def get_model(self):
        return self.model

    def set_kwargs(self, **kwargs):
        if self.listing:
            self.listing.set_kwargs(**kwargs)
        else:
            # postpone to when listing will be created
            self.store_kwargs(**kwargs)

    def set_attr(self, attr, value):
        if self.listing:
            self.listing.set_attr(attr, value)
        else:
            # postpone to when listing will be created
            self.store_kwargs(**{attr: value})

    def set_view(self, view):
        self._view = view
        if self.listing:
            self.listing.set_view(view)

    def is_initialized(self):
        if self.listing:
            return self.listing.is_initialized()
        else:
            return False

    def init(self, data, **kwargs):
        self.data = data
        if self.listing:
            self.listing.init(data, **kwargs)
        else:
            # postpone to when listing will be created
            self.store_kwargs(**kwargs)

    def create_listing(self, context):
        if not self.listing:
            request = context.request
            if self.suffix is None:
                self.suffix = Listing.get_suffix(request, self)
            variation = request.GET.get("variation" + self.suffix)
            if not variation:
                variation = request.POST.get("variation" + self.suffix)
            if not variation:
                variation = str(self.stored_params.get("variation", ""))
            if variation and variation.isdigit():
                variation = int(variation)
                if variation >= len(self.variations_classes) or variation < 0:
                    variation = self.variation_default
            else:
                variation = self.variation_default

            cls = self.variations_classes[variation]
            listing = cls(**self.init_kwargs)
            listing._view = self._view
            listing.variation = variation
            listing.variations = self
            if hasattr(self, "action"):
                listing.action = self.action
            if hasattr(self, "action_col"):
                listing.action_col = self.action_col
            listing.suffix = self.suffix
            listing.parsed_url = urlsplit(request.get_full_path())
            if self.id:
                listing.id = self.id
            # Copy variations class attributes to listing instance
            for k, v in self.__class__.__dict__.items():
                if not k.startswith("_") and not getattr(listing, k, None):
                    setattr(listing, k, v)
            listing.transfert_params_from(self)
            listing.set_kwargs()
            listing.init(self.data, context)
            self.listing = listing

    def render_init(self, context):
        self.create_listing(context)
        self.listing.set_kwargs()
        self.listing.render_init(context)

    def render_init_context(self, context):
        self.create_listing(context)
        self.listing.set_kwargs()
        self.listing.render_init_context(context)

    def render(self, context):
        self.create_listing(context)
        self.listing.set_kwargs()
        return self.listing.render(context)

    def get_url(self, context, **kwargs):
        self.create_listing(context)
        return self.listing.get_url(context, **kwargs)

    def have_to_refresh(self):
        if self.listing:
            return self.listing.have_to_refresh()
        return False

    def __setattr__(self, item, value):
        if item in LISTING_PARAMS_KEYS:
            if self.listing:
                setattr(self.listing, item, value)
                return
        super().__setattr__(item, value)

    def __getattr__(self, item):
        if self.listing and hasattr(self.listing, item):
            return getattr(self.listing, item)
        return super().__getattribute__(item)


class Listing(ListingBase):
    accept_ajax = False
    action = None
    action_button_cancel_label = pgettext_lazy("action button", "Cancel")
    action_button_edit_label = pgettext_lazy("action button", "Edit")
    action_button_update_label = pgettext_lazy("action button", "Update")
    action_button_upload_label = pgettext_lazy("action button", "Upload")
    action_col = None
    action_footer_template_name = ThemeTemplate("action_footer.html")
    action_header_template_name = ThemeTemplate("action_header.html")
    ajax_request = False
    ajax_part = None
    allow_empty_first_page = True
    anchor_hash = None
    attrs = {"class": "table table-hover table-bordered table-striped table-sm"}
    can_edit = False
    can_select = False
    chart_options_template = None
    columns = None
    columns_headers = None
    confirm_msg_for_delete = gettext_lazy(
        "Do you really want to DELETE {nb_items} item(s) ?"
    )
    confirm_msg_nb_items_for_delete = 0
    confirm_msg_for_update = gettext_lazy(
        "Do you really want to UPDATE {nb_items} items ?"
    )
    confirm_msg_nb_items_for_update = 0
    confirm_msg_for_update_all = gettext_lazy(
        "WARNING : Do you really want to UPDATE absolutely ALL {nb_all_items} items ?"
        "\n\n"
        "This will process ALL items in ALL pages in the actual listing. Note that "
        "it will take in account the filter : only filtered items will be processed."
        "\n\n"
        "IMPORTANT : Check that all form's fields are empty except those you want to "
        "update. You can use the CLEAR button to make all fields empty, then fill "
        "only one or more fields with value(s) you want to apply on all items."
    )
    confirm_msg_for_delete_all = gettext_lazy(
        "WARNING : Do you really want to DELETE absolutely ALL {nb_all_items} items ?"
        "\n\n"
        "This will process ALL items in ALL pages in the actual listing. Note that "
        "it will take in account the filter : only filtered items will be processed."
    )
    attached_form = None
    attached_form_autofill = False
    attached_form_base_class = ListingBaseForm
    attached_form_css_id = None
    container_attrs = {}
    current_page = None  # set in RecordManager.compute_current_page_records()
    data = None
    datetimepicker_date_format = "Y-m-d"
    datetimepicker_datetime_format = "Y-m-d H:i"
    datetimepicker_time_format = "H:i"
    div_template_name = ThemeTemplate("div_row.html")
    edit_on_demand = False
    editable = False
    editable_columns = set()
    editing = None
    editing_columns = None
    editing_hidden_columns = None
    editing_row_pk = None
    empty_listing_template_name = ThemeTemplate("empty_listing.html")
    empty_table_msg = None
    exclude_columns = None
    export = None
    export_columns = None
    export_max_rows = None
    exported_columns = None
    filters = None
    footer_snippet = None
    footer_template_name = None
    force_order_by = None
    force_request_data = None
    format_numbers = False
    form_model_fields = None
    form_serialize_cols = None
    form_serialize_cols_func = None
    gb_annotate_cols = None
    gb_cols = None
    gb_count_layout = "begin"  # can be "begin", "end" or None
    gb_template_name = ThemeTemplate("group_by.html")
    has_footer = False
    has_footer_action_buttons = True
    has_gb_filter_button = True
    has_group_by = False
    has_header = True
    has_hidden_selection = False
    has_nb_unfiltered_rows = False
    has_paginator = True
    has_toolbar = False
    has_upload = False
    id = None
    link_object_columns = None
    listing_template_name = ThemeTemplate("listing.html")
    model = None
    name = None
    no_permission_required_for_actions = False
    offset_max = LISTING_OFFSET_MAX
    only_one_with_prefix = None
    onready_snippet = None
    orphans = 0
    page = 1
    pagination = True
    paginator = None
    paginator_class = Paginator
    per_page = LISTING_ROWS_PER_PAGE
    per_page_max = LISTING_ROWS_PER_PAGE_MAX
    permission_required_for_export = False
    primary_key = "id"
    processed_flash = True
    processed_pks = None
    record_label = None
    record_label_plural = None
    records_class = RecordManager
    request_media_for = None
    row_attrs = {}
    row_form_base_class = ListingBaseForm
    row_form_errors = None
    save_to_database = False
    select_columns = None
    selectable = False
    selected_columns = None
    selected_hidden_columns = None
    selecting = None
    selection_initial = None
    selection_key = "id"
    selection_menu_id = None
    selection_mode = "default"  # default, overlay, hover
    selection_multiple = False
    selection_multiple_ctrl = False
    selection_overlay_template_name = ThemeTemplate("selection_overlay.html")
    selection_position = "hidden"  # left, right or hidden
    small_device_header_style = "font-weight: bold"
    sort = None
    sortable = True
    suffix = None
    toolbar = None
    toolbar_placement = "both"
    unsortable = True
    use_datetimepicker = False
    variation = None
    variations = None

    params_keys = set()  # keep it here in the class not in __init__()

    theme_class = ThemeAttribute("theme_class")
    theme_listing_class = ThemeAttribute("theme_listing_class")
    theme_action_button_class = ThemeAttribute("theme_action_button_class")
    theme_action_button_cancel_icon = ThemeAttribute("theme_action_button_cancel_icon")
    theme_action_button_edit_icon = ThemeAttribute("theme_action_button_edit_icon")
    theme_action_button_update_icon = ThemeAttribute("theme_action_button_update_icon")
    theme_action_button_upload_icon = ThemeAttribute("theme_action_button_upload_icon")
    theme_container_class = ThemeAttribute("theme_container_class")
    theme_sort_asc_icon = ThemeAttribute("theme_sort_asc_icon")
    theme_sort_desc_icon = ThemeAttribute("theme_sort_desc_icon")
    theme_sort_none_icon = ThemeAttribute("theme_sort_none_icon")
    theme_spinner_icon = ThemeAttribute("theme_spinner_icon")
    theme_sortable_class = ThemeAttribute("theme_sortable_class")
    theme_sorted_class = ThemeAttribute("theme_sorted_class")
    theme_sort_asc_class = ThemeAttribute("theme_sort_asc_class")
    theme_sort_desc_class = ThemeAttribute("theme_sort_desc_class")
    theme_button_class = ThemeAttribute("theme_button_class")
    theme_button_disabled_class = ThemeAttribute("theme_button_disabled_class")
    theme_button_active_class = ThemeAttribute("theme_button_active_class")
    theme_div_row_container_class = ThemeAttribute("theme_div_row_container_class")
    theme_row_class = ThemeAttribute("theme_row_class")
    theme_table_header_class = ThemeAttribute("theme_table_header_class")
    theme_localized_small_device_styles_width = ThemeAttribute(
        "theme_localized_small_device_styles_width"
    )

    permission_required_for_reset = False
    permission_required_for_clear = False

    def __init__(self, data=None, **kwargs):
        super().__init__(data, **kwargs)
        self.request = None
        self.request_data = None
        self.parsed_url = None
        self.page_context = None
        self.records = self.records_class(self)
        self.rows_context_list = []
        self.form_input_hiddens = {}
        self.editing_hidden_form_fields = []
        self.editing_really_hidden_columns = set()
        self.can_edit_columns = []
        self.col_cell_renderers = {}
        init_dicts_from_class(
            self, ["global_context", "attrs", "container_attrs", "row_attrs"]
        )
        self._initialized = False
        self._render_initialized = False
        self._formset = None
        self._view = None
        self.form_params = {}
        self._have_to_refresh = False
        self.columns_sort_ascending = {}
        self.columns_sort_list = []
        self.gb_filters_initial = {}

        # listing initialisation must be in 2 steps because when a class is
        # passed to a template, it is automatically instanciated by Django
        # without any parameter.
        if isinstance(data, QuerySet) or data is not None:
            self.init(data, **kwargs)

    @classmethod
    def get_params_keys(cls):
        if not Listing.params_keys:
            Listing.params_keys = set(LISTING_PARAMS_KEYS)
            Listing.params_keys.update({"columns_%s" % k for k in COLUMNS_PARAMS_KEYS})
            Listing.params_keys.update(
                {"paginator_%s" % k for k in PAGINATOR_PARAMS_KEYS}
            )
            Listing.params_keys.update({"toolbar_%s" % k for k in TOOLBAR_PARAMS_KEYS})
            Listing.params_keys.update({"filters_%s" % k for k in FILTERS_PARAMS_KEYS})
            Listing.params_keys.update(
                {"form_%s" % k for k in ATTACHED_FORM_PARAMS_KEYS}
            )
        return Listing.params_keys

    def set_kwargs(self, **kwargs):
        self.store_kwargs(**kwargs)
        params_keys = self.get_params_keys()
        for k, v in self.stored_params.items():
            if k in params_keys or ("__" in k and not k.startswith("__")):
                setattr(self, k, v)
        self.stored_params = {}

    def have_to_refresh(self):
        return self._have_to_refresh

    def plan_refresh(self):
        self._have_to_refresh = True

    def get_filtered_and_selected_queryset(self):
        qs = self.records.get_filtered_queryset()
        selected_rows = self.get_selected_rows()
        if selected_rows:
            qs = qs.filter(pk__in=selected_rows)
        return qs

    def add_onready_snippet(self, snippet):
        if not hasattr(self.request, "django_listing_onready_snippets"):
            self.request.django_listing_onready_snippets = []
        if snippet not in self.request.django_listing_onready_snippets:
            self.request.django_listing_onready_snippets.append(snippet)

    def add_header_snippet(self, snippet):
        if not hasattr(self.request, "django_listing_header_snippets"):
            self.request.django_listing_header_snippets = []
        if snippet not in self.request.django_listing_header_snippets:
            self.request.django_listing_header_snippets.append(snippet)

    def add_footer_dict_list(self, key, dct):
        if not hasattr(self.request, "django_listing_footer_dict_list"):
            self.request.django_listing_footer_dict_list = {}
        dict_list = self.request.django_listing_footer_dict_list.setdefault(key, [])
        dict_list.append(dct)

    def add_footer_snippet(self, snippet):
        if not hasattr(self.request, "django_listing_footer_snippets"):
            self.request.django_listing_footer_snippets = []
        if snippet not in self.request.django_listing_footer_snippets:
            self.request.django_listing_footer_snippets.append(snippet)

    def need_media_for(self, feature_name):
        if not self.request:
            return
        if not hasattr(self.request, "need_media_for"):
            self.request.need_media_for = {}
        self.request.need_media_for[feature_name] = True

    def create_missing_columns(self):
        if self.columns is None:
            if isinstance(self.columns_headers, str):
                self.columns_headers = self.columns_headers.split(",")
            if self.model:
                self.columns = ModelColumns(
                    self.model,
                    listing=self,
                    link_object_columns=self.link_object_columns,
                )
            elif self.data and isinstance(self.data, collections.abc.Sequence):
                self.columns = SequenceColumns(
                    self.data, self.columns_headers, listing=self
                )
        if isinstance(self.columns, (ModelColumns, SequenceColumns)):
            self.columns.set_listing(self)
            self.columns.init()

        if not self.columns:
            raise InvalidListing(
                _("Please configure at least one column " "in your listing")
            )
        if self.selectable and self.columns:
            # It is important to check selection column already exists
            # otherwise, there may be many duplicates
            if self.selection_position == "left" and not isinstance(
                self.columns[0], SelectionColumn
            ):
                self.columns.insert(0, SelectionColumn(LISTING_SELECTION_CHECKBOX_NAME))
            elif self.selection_position == "right" and not isinstance(
                self.columns[-1], SelectionColumn
            ):
                self.columns.append(SelectionColumn(LISTING_SELECTION_CHECKBOX_NAME))

    def gb_cols_dlb(self):
        select_info = []
        if self.gb_cols_names:
            for name in self.gb_cols_names:
                label = self.original_columns_headers.get(name, "-")
                select_info.append((name, label, True))
        if self.original_columns_headers:
            for name, label in self.original_columns_headers.items():
                if name not in self.gb_cols_names:
                    select_info.append((name, label, False))
        valid_names = {
            k: v["ok"]
            for k, v in validate_values_names(
                self.data, (i[0] for i in select_info)
            ).items()
        }
        valid_info = filter(lambda n: valid_names[n[0]], select_info)
        return list(sorted(valid_info, key=lambda x: x[1]))

    def gb_annotate_cols_dlb(self):
        select_info = []
        if self.gb_annotate_cols_names:
            for name in self.gb_annotate_cols_names:
                label = self.annotation_columns_headers.get(name, "-")
                select_info.append((name, label, True))
        if self.annotation_columns_headers:
            for name, label in self.annotation_columns_headers.items():
                if name not in self.gb_annotate_cols_names:
                    select_info.append((name, label, False))
        return list(sorted(select_info, key=lambda x: x[1]))

    def get_gb_listing_template_name(self):
        # str() is important here to force descriptor .get()
        return str(self.listing_template_name)

    def manage_group_by(self):
        if self.has_group_by:
            self.gb_cols_names = []
            self.gb_annotate_cols_names = []
            self.original_columns_headers = {}
            ach = self.annotation_columns_headers = {}
            for c in self.columns:
                if not c.init_args:
                    continue
                name = c.init_args[0]
                header = c.init_kwargs.get("header")
                if not header:
                    header = name.replace("_", " ").capitalize()
                if name and header:
                    self.original_columns_headers[name] = header
                if isinstance(c, (IntegerColumn, FloatColumn)):
                    for a, (alabel, afunc) in LISTING_ANNOTATIONS.items():
                        ach[f"{name}_annotate_{a}"] = f"{header} ({alabel})"
        if self.gb_cols:
            self.listing_template_name = self.get_gb_listing_template_name()
            self.attrs = {
                "class": "table table-hover table-bordered table-striped table-sm"
            }
            gb_cols_names = self.gb_cols
            if isinstance(gb_cols_names, str):
                gb_cols_names = map(str.strip, gb_cols_names.split(","))
            gb_cols_names = list(
                filter(lambda cname: cname in self.columns.name2col, gb_cols_names)
            )
            self.gb_cols_names = gb_cols_names
            gb_cols = [self.columns.get(cname) for cname in gb_cols_names]
            self.gb_filters_rec_key = {
                # equivalent to (c.attached_filter or c.name): (c.data_key or c.name)
                # but cannot do that here as columns are not totally initialized yet
                (c.init_kwargs.get("attached_filter") or c.init_args[0]): (
                    c.data_key or c.init_args[0]
                )
                for c in gb_cols
            }
            for c in gb_cols:
                # equivalent to c.attached_filter_initial
                attached_filter_initial = (
                    c.init_kwargs.get("attached_filter_initial")
                    or getattr(self, "columns_attached_filter_initial", None)
                    or {}
                )
                for k, v in attached_filter_initial.items():
                    self.gb_filters_initial[FILTER_QUERYSTRING_PREFIX + k] = v
            if self.gb_count_layout == "begin":
                gb_cols.append(IntegerColumn("count"))
            self.gb_queryset_fields = [
                # As col as not been initialized yet (and cannot do it here),
                # have to manually extract col.data_key and col.name
                (
                    getattr(self, f"{name}__gb_data_key", None)
                    or getattr(self, f"{name}__data_key", None)
                    or col.init_kwargs.get("gb_data_key")
                    or col.init_kwargs.get("data_key")
                    or col.init_args[0]
                ).replace(".", "__")
                for name in gb_cols_names
                if (col := self.columns.get(name))
            ]
            gb_annotate_cols = {"count": Count(self.gb_queryset_fields[0])}
            gb_annotate_cols_names = self.gb_annotate_cols
            if isinstance(gb_annotate_cols_names, str):
                gb_annotate_cols_names = map(
                    str.strip, gb_annotate_cols_names.split(",")
                )
            gb_annotate_cols_names = list(
                filter(lambda cname: cname in ach, gb_annotate_cols_names or [])
            )
            self.gb_annotate_cols_names = gb_annotate_cols_names
            for aname in self.gb_annotate_cols_names:
                label = self.annotation_columns_headers.get(aname)
                if label:
                    col_name, annotation = aname.split("_annotate_")
                    gb_cols.append(FloatColumn(aname, header=label))
                    gb_annotate_cols[aname] = LISTING_ANNOTATIONS[annotation][1](
                        col_name
                    )
            if self.gb_count_layout == "end":
                gb_cols.append(IntegerColumn("count"))
            if self.has_gb_filter_button:
                gb_cols.append(GroupByFilterColumn())
            self.columns = Columns(*gb_cols)
            self.select_columns = None
            self.exclude_columns = None
            self.data = self.data.values(*self.gb_queryset_fields).annotate(
                **gb_annotate_cols
            )
            if not self.sort:
                self.sort = [gb_cols_names[0]]

    def create_missing_toolbar_items(self):
        if self.toolbar is None:
            return
        if not isinstance(self, Toolbar):
            self.toolbar = Toolbar(self.toolbar)

    def create_missing_filters(self):
        if self.filters is None:
            return
        if isinstance(self.filters, str):
            # if filters attribute is a string, it means a layout was declared,
            # so we have to extract filters names
            form_layout = self.filters
            filters_name = list(
                map(
                    lambda s: s.split("|")[0].strip(),
                    sum(map(lambda s: s.split(","), form_layout.split(";")), []),
                )
            )
            self.filters = Filters(*filters_name)
            self.filters.form_layout = form_layout

    def get_model(self):
        if self.model:
            return self.model
        if self.columns:
            return self.columns.get_model()
        return None

    def is_empty(self):
        if isinstance(self.data, QuerySet):
            return self.data.exists()
        return not bool(self.data)

    def is_initialized(self):
        return self._initialized

    def is_render_initialized(self):
        return self._render_initialized

    def manage_page_context(self, context):
        if context:
            if not self.page_context:
                self.page_context = context
                self.request = context.request
                self.request_data = self.request.GET.copy()
                self.request_data.update(self.request.POST)
                if self.force_request_data:
                    self.request_data.update(self.force_request_data)
                self.parsed_url = urlsplit(self.request.get_full_path())
                self.suffix = self.get_suffix(self.request, self)
                self.extract_params()
                if self.filters:
                    self.filters.extract_params()
            else:
                self.page_context = context

    def validate_parameters(self):
        if self.editable and self.accept_ajax:
            if self.row_form_errors is None:
                self.row_form_errors = []
            self.row_form_errors.append(_("[CANNOT_EDIT_IN_AJAX]"))
            self.editable = False

    def patch_attached_form_buttons_attrs(self, button, label, attrs):
        pass

    def init(self, data, context=None, **kwargs):
        if not self.is_initialized():
            self.set_kwargs(**kwargs)
            self.validate_parameters()
            if not self.name:
                self.name = re.sub(r"([A-Z])", r"_\1", self.__class__.__name__).lower()
                self.name = self.name[1:]
            if data is None:
                data = self.model or self.columns.get_model()
            if isinstance(data, type) and issubclass(data, Model):
                self.model = data
                data = self.model.objects.all()
            elif isinstance(data, Model):
                self.model = type(data)
                data = self.model.objects.all()
            elif isinstance(data, QuerySet):
                self.model = data.model
            if self.record_label is None:
                if self.model:
                    self.record_label = self.model._meta.verbose_name
                else:
                    self.record_label = _("record")
            if self.record_label_plural is None:
                if self.model:
                    self.record_label_plural = self.model._meta.verbose_name_plural
                else:
                    self.record_label_plural = _("records")
            if self.empty_table_msg is None:
                self.empty_table_msg = (
                    _("There is no %s to display.") % self.record_label
                )

            self.data = data
            self.create_missing_filters()
            if self.filters:
                self.filters = self.filters.bind_to_listing(self)
            self.manage_page_context(context)
            self.create_missing_columns()
            self._initialized = True

    def datetimepicker_init(self):
        if self.use_datetimepicker:
            self.need_media_for("datetimepicker")
            self.add_footer_dict_list(
                "datetimepickers", dict(listing=self, div_id=self.id)
            )

    def dropzone_init(self):
        self.need_media_for("dropzone")
        dz_suffix = self.suffix[1:] if isinstance(self.suffix, str) else ""
        dz_camel_name = f"actionForm{dz_suffix}"
        self.add_footer_dict_list(
            "dropzones",
            dict(
                listing=self,
                dz_camel_name=dz_camel_name,
                options=settings.django_listing_settings.DROPZONE_PARAMS,
            ),
        )

    def global_context_init(self):
        self.global_context.update(settings.django_listing_settings.context)
        if self.request and (self.can_edit or self.has_upload):
            self.global_context["csrf_token"] = get_csrf_token(self.request)

    def render_init_context(self, context):
        self.manage_page_context(context)
        self.normalize_params()
        self.add_form_input_hiddens(requested_url=self.request.get_full_path())
        self.manage_group_by()
        self.create_missing_toolbar_items()
        self.has_toolbar = bool(self.toolbar)
        self.columns = self.columns.bind_to_listing(self)
        for col in self.columns:
            col.render_init_context(context)
        if self.attached_form:
            self.attached_form = self.attached_form.bind_to_listing(self)
        if self.toolbar:
            self.toolbar = self.toolbar.bind_to_listing(self)
        self.col_cell_renderers = {
            col: getattr(self, "render_{}".format(col.name), col.render_cell)
            for col in self.columns
        }
        if not isinstance(self.attrs, HTMLAttributes):
            self.attrs = HTMLAttributes(self.attrs)
        html_class = "listing-" + self.__class__.__name__.lower()
        self.attrs.add("class", {html_class, self.theme_listing_class})
        if self.variation is not None:
            self.attrs.add("class", "variation-{}".format(self.variation))
        if not self.id:
            self.id = "{}{}-id".format(html_class, self.suffix)
        self.css_id = self.id.replace("_", "-")
        if isinstance(self.suffix, str):
            self.css_suffix = self.suffix.replace("_", "-")
        if isinstance(self.editable_columns, str):
            self.editable_columns = set(
                map(str.strip, self.editable_columns.split(","))
            )
        if self.editing_hidden_columns is None:
            if self.model and self.primary_key in self.columns.names():
                self.editing_hidden_columns = {self.primary_key}
            else:
                self.editing_hidden_columns = set()
        if isinstance(self.editing_hidden_columns, str):
            self.editing_hidden_columns = set(
                map(str.strip, self.editing_hidden_columns.split(","))
            )
        if self.editing is None:
            self.editing = True
        if self.editing_columns is None:
            self.editing_columns = "all"
        if isinstance(self.editing, str) and self.editing.lower() == "false":
            self.editing = False
        if isinstance(self.editing_columns, str):
            self.editing_columns = set(map(str.strip, self.editing_columns.split(",")))
        self.columns.editing_init()
        self.can_edit = self.editable and self.editing
        if isinstance(self.selecting, str) and self.selecting.lower() == "false":
            self.selecting = False
        if self.selecting is None:
            self.selecting = True
        self.can_select = self.selectable and self.selecting
        self.has_hidden_selection = (
            self.selectable and self.selecting and self.selection_position == "hidden"
        )
        self.columns_sort_ascending = {}
        self.columns_sort_list = []
        if self.sort:
            if isinstance(self.sort, str):
                sort_list = map(str.strip, self.sort.split(","))
            else:
                sort_list = self.sort
            for col_name in sort_list:
                ascending = True
                if col_name.startswith("-"):
                    col_name = col_name[1:]
                    ascending = False
                self.columns_sort_list.append(col_name)
                self.columns_sort_ascending[col_name] = ascending
        if not isinstance(self.row_attrs, HTMLAttributes):
            self.row_attrs = HTMLAttributes(self.row_attrs)
        if not isinstance(self.container_attrs, HTMLAttributes):
            self.container_attrs = HTMLAttributes(self.container_attrs)
        self.row_attrs.add(
            "class", {self.theme_div_row_container_class, self.theme_row_class}
        )
        self.selected_columns = self.columns.select(
            self.select_columns, self.exclude_columns
        )
        self.exported_columns = self.columns.select_exported(self.export_columns)
        if self.exported_columns is None:
            self.exported_columns = self.selected_columns
        self.can_edit_columns = [c for c in self.selected_columns if c.can_edit]
        self.can_edit_columns_names = set(c.name for c in self.can_edit_columns)
        self.editing_really_hidden_columns = (
            self.editing_hidden_columns - self.can_edit_columns_names
        )
        self.selected_hidden_columns = self.columns.select(
            self.editing_really_hidden_columns
        )
        if isinstance(self.request_media_for, str):
            self.request_media_for = map(str.strip, self.request_media_for.split(","))

    def render_init(self, context):
        if not self._render_initialized:
            self.render_init_context(context)
            for col in self.columns:
                col.render_init()
            if self.request_media_for:
                for feature in self.request_media_for:
                    self.need_media_for(feature)
            try:
                is_exporting = self.export_data()
                if is_exporting:
                    return "Sending listing export file..."
            except ListingException as e:
                messages.error(self.request, str(e))
            if self.can_edit:
                self.datetimepicker_init()
            if self.has_upload:
                self.dropzone_init()
            if (
                self.can_edit
                or self.can_select
                or self.attached_form
                or self.has_upload
            ):
                self.add_form_input_hiddens(
                    listing_id=self.id, listing_suffix=self.suffix
                )
            if self.onready_snippet:
                self.add_onready_snippet(self.onready_snippet)
            if self.footer_snippet:
                self.add_footer_snippet(self.footer_snippet)
            if self.selection_initial is not None and not isinstance(
                self.selection_initial, list
            ):
                self.selection_initial = [self.selection_initial]
            if self.attached_form:
                self.attached_form.render_init(context)
            self.global_context_init()
            self.do_filter_data()
            self.compute_current_page_records()
            self._render_initialized = True

    def do_filter_data(self):
        # Last chance to filter data before listing page computation
        pass

    def compute_current_page_records(self):
        """Compute the current page records against the actual given data

        This method is to be called if the listing data has changed after initialization.
        This is used in attached form POST request management.
        """
        self.records.compute_current_page_records()

    def render(self, context):
        response = self.render_init(context)
        if response is not None:
            return response
        return self.render_template()

    def export_data(self):
        if self.export and not hasattr(self.request, "export_data"):
            if not self.has_permission_for_action("export"):
                raise ListingException(_("[You do not have permission to export data]"))
            if self.export_max_rows and self.exported_nb_rows() > self.export_max_rows:
                raise ListingException(
                    _(
                        "[You cannot export when there are more than {} rows in listing]"
                    ).format(self.export_max_rows)
                )
            export_format = self.export.upper()
            if export_format in EXPORT_FORMATS:
                use_col_name = export_format in EXPORT_FORMATS_USE_COL_NAME
                headers = self.exported_headers(use_col_name)
                if export_format == "DBF":
                    headers = list(map(lambda h: h[:10], headers))
                data = tablib.Dataset(title=_("[Listing export]"))
                data.headers = headers
                keep_original_type = export_format in EXPORT_FORMATS_KEEP_ORIGINAL_TYPE
                for row in self.exported_rows(keep_original_type):
                    data.append(row)
                export_params = {}
                if export_format == "CSV":
                    export_params["delimiter"] = ";"
                self.request.export_data = data.export(
                    export_format.lower(), **export_params
                )
                current_date = timezone.now().strftime("%Y-%m-%d.%Hh%M")
                self.request.export_filename = "{}.{}.{}".format(
                    self.name, current_date, self.export.lower()
                )
        return hasattr(self.request, "export_data")

    def django_listing_info(self):
        if hasattr(self, "request"):
            if self.request.GET.get("__django_listing_info__"):
                out = "<b><u>django-listing informations :</u></b><br>"
                import django_listing

                out += f"django-listing version : {django_listing.__version__}<br>"
                import django

                out += f"django version : {django.__version__}<br>"
                import tablib

                out += f"tablib version : {tablib.__version__}<br>"
                import sys

                out += f"Python version : {sys.version.split()[0]}<br>"
                return out

    def render_template(self):
        request_for_info = self.django_listing_info()
        if request_for_info:
            return request_for_info
        ctx = self.get_listing_context()
        template = loader.get_template(self.listing_template_name)
        # request is needed for context_processor processing
        out = template.render(ctx, request=self.request)
        return out

    def get_listing_css_id(self):
        return str(self.id).replace("_", "-")

    def get_listing_context(self):
        has_top_toolbar = self.has_toolbar and self.toolbar_placement in ["top", "both"]
        has_bottom_toolbar = self.has_toolbar and self.toolbar_placement in [
            "bottom",
            "both",
        ]
        self.selection_has_overlay = self.selection_mode in ["overlay", "hover"]
        ca = self.container_attrs  # This is a HtmlAttributes object
        ca.add("id", self.css_id)
        ca.add("class", {self.theme_container_class, self.theme_class})
        if self.accept_ajax:
            ca.add("class", "django-listing-ajax")
            ca.add("class", "django-listing-ajax")
        if self.can_edit:
            ca.add("class", "django-listing-editing")
        if self.can_select:
            ca.add("class", "django-listing-selecting")
            if self.selection_multiple:
                ca.add("class", "selection_multiple")
                if self.selection_multiple_ctrl:
                    ca.add("class", "selection-multiple-ctrl")
            else:
                ca.add("class", "selection_unique")
            ca.add("class", f"selection_position_{self.selection_position}")
        if self.has_upload:
            ca.add("class", "has_upload")
        if self.attached_form_autofill:
            ca.add("class", "attached_form_autofill")
        if self.format_numbers:
            ca.add("class", "format-numbers")
        if self.gb_cols:
            ca.add("class", "has-gb-cols")
        ca.add("ajax_url", self.get_url())
        if self.selection_menu_id:
            ca.add("selection-menu-id", self.selection_menu_id)
        ca.add("attached-form-id", self.attached_form_css_id)
        if self.paginator:
            ca.add("nb-rows", self.paginator.count)
        if self.has_nb_unfiltered_rows:
            ca.add("nb-unfiltered-rows", self.records.get_unfiltered_count())
        sel_css_class = LISTING_SELECTOR_CSS_CLASS if self.selection_has_overlay else ""
        hover_css_class = (
            LISTING_SELECTION_HOVER_CSS_CLASS if self.selection_mode == "hover" else ""
        )
        ctx = RenderContext(
            self.global_context,
            self.page_context.flatten(),
            nb_columns=len(self.selected_columns),
            listing=self,
            has_top_toolbar=has_top_toolbar,
            has_bottom_toolbar=has_bottom_toolbar,
            get=self.request.GET,
            post=self.request.POST,
            overlay_selector_css_class=sel_css_class,
            hover_selection_css_class=hover_css_class,
            view=self.get_view(),
            # Needed for action framework with javascript compatibility :
            csrf_token=get_csrf_token(self.request),
        )
        if self.paginator:
            ctx.update(self.paginator.get_context())
        if self.toolbar:
            ctx.update(self.toolbar.get_context())
        return ctx

    def add_form_input_hiddens(self, **kwargs):
        self.form_input_hiddens.update(kwargs)

    def add_edit_form_hidden_field(self, field_instance):
        self.editing_hidden_form_fields.append(field_instance)

    def aggregate(self, rec):
        for col in self.selected_columns:
            if col.aggregation:
                col.aggregation.aggregate(rec)

        if self.can_edit:
            for col in self.selected_hidden_columns:
                form = rec.get_form()
                if form:
                    self.add_edit_form_hidden_field(form[col.name])

    def update_page_records(self, records):
        method = getattr(self._view, "update_page_records", None)
        if callable(method):
            method(self, records)

    def div_rows(self):
        for rec in self.records.current_page():
            self.rows_context_list.append(rec)
            index = rec.get_index()
            attrs = self.get_row_attrs(rec)  # must be executed before creating the dict
            row = dict(
                attrs=attrs,
                index=index,
                index1=index + 1,
                selected=rec.is_selected(),
                rec=rec,
            )
            if self.has_hidden_selection:
                row.update(selection_value=rec.get(self.selection_key))
            self.aggregate(rec)
            yield row

    def rows(self):
        for rec in self.records.current_page():
            index = rec.get_index()
            attrs = self.get_row_attrs(rec)  # must be executed before creating the dict
            row = dict(
                attrs=attrs,
                ctx=self.get_row_context(rec),
                columns=self.get_rendered_cells(rec),
                index=index,
                index1=index + 1,
                selected=rec.is_selected(),
                rec=rec,
            )
            if self.has_hidden_selection:
                row.update(selection_value=rec.get(self.selection_key))
            self.aggregate(rec)
            yield row

    def exported_headers(self, use_col_name=True):
        if use_col_name:
            return [c.name for c in self.exported_columns if c.exportable]
        else:
            return [
                str(c.get_exported_header_value())
                for c in self.exported_columns
                if c.exportable
            ]

    def get_cell_exported_value(self, col, rec, keep_original_type):
        cell_value_func = getattr(self, f"exported_cell_value_{col.name}", None)
        if cell_value_func:
            value = cell_value_func(rec, keep_original_type)
        else:
            value = col.get_cell_exported_value(rec, keep_original_type)
        # Excel sanitization
        if isinstance(value, str):
            value = EXPORT_EXCEL_SANITIZE_RE.sub(" ", value)
        elif isinstance(value, datetime):
            if timezone.is_aware(value):
                # Convert to the same timezone as the input datetime
                # and then remove the timezone information
                return timezone.localtime(value).replace(tzinfo=None)
            value = value.replace(tzinfo=None)
        return value

    def exported_nb_rows(self):
        return self.records.export_count()

    def exported_rows(self, keep_original_type=True):
        for rec in self.records.export():
            yield [
                self.get_cell_exported_value(c, rec, keep_original_type)
                for c in self.exported_columns
                if c.exportable
            ]

    def get_rendered_cells(self, rec):
        rendered_columns = []
        for col in self.selected_columns:
            rendered_col = dict(
                html=self.col_cell_renderers[col](rec),
                obj=col,
            )
            rendered_columns.append(rendered_col)
        return rendered_columns

    def header_columns(self):
        if self.has_header:
            rendered_columns = []
            for col in self.selected_columns:
                rendered_col = dict(
                    html=col.render_header(),
                    obj=col,
                )
                rendered_columns.append(rendered_col)
            return rendered_columns
        else:
            return []

    def get_localized_small_device_styles(self):
        styles = [
            ("#{listing_css_id} th:before {{" 'content:"{sort_msg}";' "}}").format(
                listing_css_id=self.get_listing_css_id(),
                sort_msg=_("Sort by:"),
                two_dots=_(":"),
            )
        ]
        for col in self.selected_columns:
            style = (
                "#{listing_css_id} td.col-{col_name}:before {{"
                'content:"{header}{two_dots} ";'
                "{sd_style};"
                "}}"
            ).format(
                listing_css_id=self.get_listing_css_id(),
                col_name=col.name,
                header=col.get_header_value(),
                two_dots=_(":"),
                sd_style=self.small_device_header_style,
            )
            styles.append(style)
        return styles

    def footer_columns(self):
        if self.has_footer:
            rendered_columns = []
            for col in self.selected_columns:
                rendered_col = dict(
                    html=col.render_footer(),
                    obj=col,
                )
                rendered_columns.append(rendered_col)
            return rendered_columns
        else:
            return []

    def get_url(
        self,
        context=None,
        without=None,
        anchor_hash=None,
        only_one_with_prefix=None,
        **kwargs,
    ):
        """Get listing url with some updated parameters if needed
        Note : Do not remove 'context=None' because needed by listing variations
        """
        querystring = QueryDict(self.parsed_url.query, mutable=True)
        if only_one_with_prefix:
            del_keys = [
                k for k in querystring.keys() if k.startswith(only_one_with_prefix)
            ]
            for k in del_keys:
                del querystring[k]
        for k, v in kwargs.items():
            if not isinstance(v, str):
                v = str(v)
            if k in LISTING_QUERY_STRING_KEYS:
                k += self.suffix
            querystring[k] = v
        del_keys = LISTING_NOT_PERSISTENT_QUERY_STRING_KEYS - set(kwargs.keys())
        if without:
            if isinstance(without, str):
                without = without.split(",")
            del_keys = del_keys | set(without)
        for k in del_keys:
            k += self.suffix
            if k in querystring:
                del querystring[k]
        fragment = anchor_hash or self.anchor_hash
        return urlunsplit(
            self.parsed_url._replace(query=querystring.urlencode(), fragment=fragment)
        )

    def get_hiddens(self, without=None):
        hiddens = QueryDict(self.parsed_url.query, mutable=True)
        if self.filters and "f_do_filter" not in hiddens:
            initial = self.filters.form().initial
            if initial:
                hiddens.update(initial)
        if without:
            if isinstance(without, str):
                without = list(map(str.strip, without.split(",")))
            without = list(map(lambda s: s.strip() + self.suffix, without))
            if isinstance(without, (list, tuple)):
                for k in without:
                    if k in hiddens:
                        del hiddens[k]
        return hiddens

    def get_hiddens_html(self, without=None, **additional_hiddens):
        hiddens = self.get_hiddens(without)
        out = [
            '<input type="hidden" name="{name}" value="{value}">'.format(
                name=keylist[0], value=value
            )
            for keylist in hiddens.lists()
            for value in keylist[1]
        ]
        for k, v in additional_hiddens.items():
            out.append(f'<input type="hidden" name="{k}" value="{v}">')
        return "".join(out)

    @classmethod
    def get_suffix_request_data(cls, request):
        if not hasattr(request, LISTING_SUFFIX_REQUEST_DATA_FIELD):
            setattr(
                request,
                LISTING_SUFFIX_REQUEST_DATA_FIELD,
                {
                    "listings_suffixes": {},
                    "counter": 0,
                },
            )
        return getattr(request, LISTING_SUFFIX_REQUEST_DATA_FIELD)

    @classmethod
    def get_suffix(cls, request, listing):
        listing = getattr(listing, "variations", None) or listing
        data = cls.get_suffix_request_data(request)
        suffix = data["listings_suffixes"].get(listing)
        if suffix is None:
            counter = data["counter"]
            suffixes = data["listings_suffixes"].values()
            while True:
                if counter == 0:
                    suffix = ""
                else:
                    suffix = LISTING_SUFFIX_PATTERN.format(counter)
                if suffix not in suffixes:
                    data["listings_suffixes"][listing] = suffix
                    break
                counter += 1
            data["counter"] = counter + 1
        return suffix

    @classmethod
    def set_suffix(cls, request, listing, suffix):
        listing = getattr(listing, "variations", None) or listing
        data = cls.get_suffix_request_data(request)
        data["listings_suffixes"][listing] = suffix
        listing.suffix = suffix

    def extract_params(self):
        for k in LISTING_QUERY_STRING_KEYS:
            qs_key = k + self.suffix
            v = None
            if qs_key in self.request_data:
                v = self.request_data.get(qs_key)
            if v is not None and k in LISTING_QUERY_STRING_INT_KEYS:
                try:
                    v = int(v)
                except ValueError:
                    pass
            if v is not None:
                setattr(self, k, v)

    def normalize_params(self):
        if self.per_page < -1 or self.per_page == 0:
            self.per_page = self.__class__.per_page

    def get_row_attrs(self, rec):
        attrs = HTMLAttributes(self.row_attrs)  # create a copy
        if rec.pk:
            attrs.add("data-pk", str(rec.pk))
        if self.form_model_fields and self.attached_form_autofill and not self.gb_cols:
            attrs.add(
                "data-serialized-object",
                rec.get_serialized_object(fields=self.form_model_fields),
            )
        if self.processed_flash:
            if self.processed_pks and rec.pk in self.processed_pks:
                attrs.add("class", {"flash-once", "selected"})
                attrs.add("class", {self.action.replace("_", "-")})
                attrs.add("class", {self.action_button.replace("_", "-")})
        attrs.add("class", "odd" if rec.get_index() % 2 else "even")
        if rec.is_first_qs_record():
            attrs.add("class", "qs-first")
        if rec.is_last_qs_record():
            attrs.add("class", "qs-last")
        if self.can_select:
            if not self.selection_has_overlay:
                attrs.add("class", LISTING_SELECTOR_CSS_CLASS)
            selection_value = rec.get(self.selection_key)
            if self.selection_initial and selection_value in self.selection_initial:
                attrs.add("class", "selected")
                rec.set_selected()
        return attrs

    def get_row_context(self, rec):
        return {}

    def get_formset_initial_values(self):
        initial_values = [
            {
                c.name: c.get_cell_form_value(rec)
                for c in (self.can_edit_columns + self.selected_hidden_columns)
            }
            for rec in self.records.current_page()
            if not self.editing_row_pk or rec.pk == self.editing_row_pk
        ]
        return initial_values

    def get_formset(self):
        if not self._formset:
            post_data = None
            post_files = None
            if self.request.method == "POST":
                posted_listing_id = self.request.POST.get("listing_id", "")
                if posted_listing_id == self.id:
                    post_data = self.request.POST
                    post_files = self.request.FILES
            fields = {
                c.name: c.create_hidden_form_field()
                for c in self.selected_hidden_columns
            }
            fields.update(
                {c.name: c.create_form_field() for c in self.can_edit_columns}
            )
            form_class = type(
                "ListingRowForm{}".format(self.suffix),
                (self.row_form_base_class,),
                {"base_fields": fields},
            )
            formset_class = forms.formset_factory(form_class, extra=0)
            self._formset = formset_class(
                post_data,
                post_files,
                initial=self.get_formset_initial_values(),
                prefix="{}{}".format(attached_formset_prefix, self.suffix),
            )
            for form in self._formset:
                form.listing = self
                form.form_name = "row_form"
        return self._formset

    def get_selected_rows(self):
        if not hasattr(self, "_selected_rows"):
            self._selected_rows = []
            if self.request and self.request.POST:
                # Read selected listing rows
                if "selected_pks" in self.request.POST:
                    self._selected_rows = self.request.POST.get("selected_pks", [])
                    if isinstance(self._selected_rows, str):
                        self._selected_rows = list(
                            map(
                                lambda x: int(x),
                                filter(
                                    str.isdigit,
                                    map(str.strip, self._selected_rows.split(",")),
                                ),
                            )
                        )
                else:
                    self._selected_rows = (
                        self.request.POST.getlist(
                            LISTING_SELECTION_INPUT_NAME_KEY + self.suffix
                        )
                        or []
                    )
        return self._selected_rows

    def set_view(self, view):
        self._view = view

    def get_view(self):
        return self._view

    def get_confirm_msg_for_action(self, action):
        attr_name = f"confirm_msg_for_{action}"
        msg = getattr(self, attr_name, "")
        return msg

    def get_confirm_msg_nb_items_for_action(self, action):
        attr_name = f"confirm_msg_nb_items_for_{action}"
        nb = getattr(self, attr_name, "")
        return nb

    def has_permission_for_action(self, action):
        if not action:
            raise ListingException("self.action_button value is empty")
        no_permission_required = getattr(
            self._view, "no_permission_required_for_actions", None
        )
        if no_permission_required is None:
            no_permission_required = getattr(self, "no_permission_required_for_actions")
        if no_permission_required:
            return True
        perm_attr_name = f"permission_required_for_{action}"
        perms = getattr(self._view, perm_attr_name, None)
        if perms is None:
            perms = getattr(self, perm_attr_name, True)
        if perms is True:
            model = self.model or getattr(self._view, "model", None)
            if model:
                app_label = model._meta.app_label
                model_name = model._meta.model_name
                permission_action = self.PERMISSION_ACTIONS.get(action, action)
                if permission_action:
                    perms = (f"{app_label}.{permission_action}_{model_name}",)
        if perms == False:
            return True
        if perms is not None:
            return self.request.user.has_perms(perms)
        return True


class DivListing(Listing):
    listing_template_name = ThemeTemplate("listing_div.html")
    theme_row_class = "row-container"

    def get_gb_listing_template_name(self):
        # str() is important here to force descriptor .get()
        # div is unwanted when grouping : go back to standard listing
        return str(ThemeTemplate("listing.html"))
